////////////////////////////////////////////////////////////////
// Copyright (C) 1992, 1993, 1994 Sheetal Kakkad, Donovan Kolbly and
// Paul Wilson
//
// This file is part of the Texas Persistent Storage system.
//
// The Texas Persistent Storage system is free software; you can
// redistribute it and/or modify it under the terms of the GNU General
// Public License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// The Texas Persistent Storage system is distributed in the hope that
// it will be useful, but WITHOUT ANY WARRANTY; without even the
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
// PURPOSE. See the GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with the Texas Persistent Storage system; if not, write to
// the Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA
// 02139, USA.
//
// You can contact the authors by email at <oops@cs.utexas.edu>.
//

////////////////////////////////////////////////////////////////
// $Id: decode.cc,v 1.1 1995/02/16 18:21:42 steve Exp $
//
// Decoders are Swizzler member functions which know how to interpret
// fast type descriptors to swizzler and unswizzle pages.
//

#include <errno.h>
#include <string.h>
#include "errors.hh"
#include "pubtypes.hh"
#include "malloc.hh"
#include "objstore.hh"
#include "roottbl.hh"
#include "multpage.hh"
#include "swizzler.hh"
#include "vfthash.hh"
#include "swizzmap.hh"
#include <assert.h>



////////////////////////////////////////////////////////////////
// Swizzler::unswizzle_vft_pointer
//
// Unswizzles the virtual function table pointer. This is done by
// looking up the address of the vft in the hash table, and getting
// it's corresponding name and index in vft table associated with the
// persistent store i.e. table stored in file "<pstore>.vft". If the
// name doesn't exist in the vft table, it is added to it. Finally,
// the index is substituted for the address to unswizzle the vft
// pointer.
//

void Swizzler::unswizzle_vft_pointer (VirtAddress ptrp)
{
  int index;
  char *name = NULL;
  long addr = *((long *) ptrp);

  index = pvmap->ps->vft_hash_table->get_vft_index (addr, &name);
  if (index == INVALID_INDEX)
  {
    // If the index is not found, and no name is found associated with
    // the given address, then it's a serious problem, because we
    // cannot put the index in the vft position (i.e. cannot
    // unswizzle the vft pointer).
    if (name == NULL)
      FATAL_ERROR (BAD_VFT_PTR);

    // Otherwise, add the name to the VFT table and install the
    // mapping from the name to its index in the hash table.
    char *vft_name = new char[strlen(name)+1];
    strcpy (vft_name, name);
    VFTEntry *entry = new VFTEntry (vft_name, INVALID_ADDRESS);
    index = pvmap->ps->vft_table->add_element (entry);
    pvmap->ps->vft_hash_table->set_vft_index (name, index);
  }
  
  // Replace the vft ptr with the vft index.
  *((int *) ptrp) = index;
}



////////////////////////////////////////////////////////////////
// Swizzler::swizzle_vft_pointer
//
// Swizzle the virtual function table pointer. This is essentially the
// opposite of the "unswizzle_vft_pointer" function, since it uses
// the index to get the name, looks up the name in the hash table to
// get the address and installs the address in the vft pointer
// position.
//

void Swizzler::swizzle_vft_pointer (VirtAddress ptrp)
{
  int index;
  char *name;
  long addr;

  // Extract the inde.
  index = *((int *) ptrp);

  // Find the name corresponding to the index from the VFT table.
  name = pvmap->ps->vft_table->get_name (index);

  // Look up the name in the hash table, to get the address of the
  // virtual function table in the executable. This value cannot be
  // invalid, because that means that the executable does not know
  // about this virtual function (and the class) and we can't swizzle
  // it.
  addr = pvmap->ps->vft_hash_table->get_vft_addr (name);
  assert (addr != INVALID_ADDRESS);

  // Replace the index with the address of the vft.
  *((long *) ptrp) = addr;
}



////////////////////////////////////////////////////////////////
// Swizzler::convert_pointer
//
// Based on the swizzling direction, swizzle or unswizzle the pointer.
//

inline void Swizzler::convert_pointer (SwizzleDirection dir, VirtAddress ptrf)
{
  if (dir == PA2VA)
    swizzle_pointer ((PersAddress *) ptrf);
  else
    unswizzle_pointer ((VirtAddress *) ptrf);
}



////////////////////////////////////////////////////////////////
// Swizzler::field_of_interest
//
// Take action for the specified field based on the type of the field.
// This can be either a regular pointer, a smart pointer (when we
// implement them) or a vft pointer. In each case, call the
// corresponding routine to do the swizzling/unswizzling based on the
// swizzling direction.
//

inline void Swizzler::field_of_interest (VirtAddress obj_base,
					 InstanceField& fld, 
					 SwizzleDirection dir)
{
  VirtAddress ptrp = obj_base + fld.offset;

  switch (fld.mode)
  {
   case SWIZZLE_POINTER:
    convert_pointer (dir, ptrp);
    break;

   case SWIZZLE_SMART_POINTER:
    // FIXME. We don't have general swizzled smart pointers fully
    // implemented yet. Do we ever want to?
#if 0
    if (dir == VA2PA)
      *(unsigned long *) ptrp = INVALID_ADDRESS;
#endif
    assert (0);
    break;

   case SWIZZLE_VFT_POINTER:
    if (dir == PA2VA)
      swizzle_vft_pointer (ptrp);
    else
      unswizzle_vft_pointer (ptrp);
    break;

   default:
    // Should never reach here...
    assert (0);
    break;
  }
}



////////////////////////////////////////////////////////////////
// Swizzler::decode_offsets_wholly_contained
//
// Decode offsets specified by the interesting fields in the "part",
// where this part is *wholly* contained inside the region being
// decoded. The second argument specifies the virtual address of the
// beginning of the part. See figure in file "largeobj.ps" in "docs"
// directory.
//

inline void Swizzler::decode_offsets_wholly_contained (FastTDescPart *part,
						       VirtAddress va,
						       SwizzleDirection dir)
{
#ifdef VERBOSE_DECODING
  printf ("part [%#x - %#x] is wholly contained\n", 
	  va, va + part->size_of_area_of_interest);
#endif

  // Loop through all the interesting fields in the part, decoding
  // each field appropriately.
  for (unsigned int i = 0; i < part->num_interesting_fields; i++)
    field_of_interest (va, part->interesting_fields[i], dir);
}



////////////////////////////////////////////////////////////////
// Swizzler::decode_offsets_initial_contained
//
// Decode offsets specified by the interesting fields in the "part",
// where only some of the *initial* interesting fields are contained
// inside the region being decoded. The second argument specifies the
// virtual address of the beginning of the part, while the third
// argument specifies the end of the decoding region. See figure in
// file "largeobj.ps" in "docs" directory.
//

inline void Swizzler::decode_offsets_initial_contained (FastTDescPart *part,
							VirtAddress va,
							VirtAddress end_region,
							SwizzleDirection dir)
{
#ifdef VERBOSE_DECODING
  printf ("part [%#x - %#x] is initially contained\n", 
	  va, va + part->size_of_area_of_interest);
#endif

  for (unsigned int i = 0; i < part->num_interesting_fields; i++)
  {
    // If the field lies beyond the region of decoding, then we are
    // done and we can break the loop (and return).
    if ((va + part->interesting_fields[i].offset) >= end_region)
      break;

    // Otherwise, decode the field.
    field_of_interest (va, part->interesting_fields[i], dir);
  }
}



////////////////////////////////////////////////////////////////
// Swizzler::decode_offsets_final_contained
//
// Decode offsets specified by the interesting fields in the "part",
// where only some of the *final* interesting fields are contained
// inside the region being decoded. The second argument specifes the
// virtual address of the beginning of the part, while the third
// argument specifies the beginning of the decoding region. See figure
// in file "largeobj.ps" in "docs" directory.
//

inline void Swizzler::decode_offsets_final_contained (FastTDescPart *part,
						      VirtAddress va,
						      VirtAddress beg_region,
						      SwizzleDirection dir)
{
#ifdef VERBOSE_DECODING
  printf ("part [%#x - %#x] is finally contained\n", 
	  part_start_va, part_start_va + part->size_of_area_of_interest);
#endif

  unsigned int i = 0;

  // First, skip over the initial offsets which lie outside the region
  // being decoded.
  while ((i < part->num_interesting_fields) && 
	 ((va + part->interesting_fields[i].offset) < beg_region))
    i++;

  // Now, decode each of the remaining final fields appropriately.
  while (i < part->num_interesting_fields)
  {
    field_of_interest (va, part->interesting_fields[i], dir);
    i++;
  }
}



////////////////////////////////////////////////////////////////
// Swizzler::decode_offsets_middle_contained
//
// Decode offsets specified by the interesting fields in the "part",
// where only some of the *middle* interesting fields are contained
// inside the region being decoded. The second argument specifes the
// virtual address of the beginning of the part, the third argument
// specifies the beginning of the decoding region and the fourth
// argument specifies the end of the decoding region. This method is
// essentially the "union" of the previous two methods. See figure in
// file "largeobj.ps" in "docs" directory.
// 

inline void Swizzler::decode_offsets_middle_contained (FastTDescPart *part,
						       VirtAddress va,
						       VirtAddress beg_region, 
						       VirtAddress end_region,
						       SwizzleDirection dir)
{
#ifdef VERBOSE_DECODING
  printf ("not contained within [%#x - %#x]\n", 
	  part_start_va, part_start_va + part->size_of_area_of_interest);
#endif
  unsigned int i = 0;

  // First, skip over the initial offsets which lie outside (before)
  // the region being decoded.
  while ((i < part->num_interesting_fields) &&
	 (va + part->interesting_fields[i].offset) < beg_region)
    i++;

  // Now, decode each of the remaining final fields appropriately,
  // checking for region boundary.
  while (i < part->num_interesting_fields)
  {
    // If the field lies beyond (after) the region of decoding, then
    // we are done and we can break the loop (and return).
    if ((va + part->interesting_fields[i].offset) >= end_region)
      break;

    // Otherwise decode the field.
    field_of_interest (va, part->interesting_fields[i], dir);

    i++;
  }
}



////////////////////////////////////////////////////////////////
// Swizzler::decode_small_objects_page
//
// Starting at the given address, translate all pointers in a page.
// This function works only with pages that contain entire objects
// (i.e. objects not larger than a page). For pages that are in the
// middle of large (multi-page) objects, use
// Swizzler::decode_large_object_page.
//

void Swizzler::decode_small_objects_page (VirtAddress va, PersAddress pa,
					  SwizzleDirection dir)
{
  size_t chunk_size;

#ifdef VERBOSE_DECODING
  printf ("Decoding page: va %#x, pa %#x, dir %s\n", va, pa, 
	  (dir == VA2PA) ? "v->p" : "p->v");
#endif

  // Sanity check; `va' should be the virtual memory address
  // corresponding to `pa'
  assert (va == pvmap->get_virt_address (pa));

  // The chunk size of each chunk on the page is the same, and is
  // given by 2^size_class of any object on the page.
  chunk_size = 1 << ((MemBlockHdr *) va)->get_size_class ();

  // Since we are only swizzling one page, the zone end is one page
  // away.
  VirtAddress zone_end = va + HEAP_PAGE_SIZE;

  // For each chunk in the page, swizzle the object in the chunk if
  // the chunk is allocated, or the "next" pointer in the header if
  // the chunk is free.
  for (VirtAddress ptr = va; ptr < zone_end; ptr += chunk_size)
  {
    MemBlockHdr *hunk = (MemBlockHdr *) ptr;

    // Sanity check -- each chunk should be of the same size on a
    // given page (in other words, each chunk should have the same
    // size class).
    assert (chunk_size == (1 << hunk->get_size_class()));

    if (hunk->get_alloc_flag() == ISFREE)
    {
      // This chunk is free, and so the only pointer is the "next"
      // field in the header used for stringing free chunks together.
      convert_pointer (dir, (VirtAddress) &hunk->next);
      num_next_ptrs_swizzled++;
    }
    else
    {
      // The chunk is allocated, and we have to swizzle the object
      // based on its type descriptor.

      assert (hunk->get_alloc_flag () == ISALLOC ||
	      hunk->get_alloc_flag () == ISMEMALIGN);

      // Skip past the header to get pointer to the actual beginning
      // of the object.
      VirtAddress obj = ptr + sizeof (MemBlockHdr);
      PStore *ps = pvmap->ps;
      unsigned long obj_size;
      unsigned int type_id;
      Boolean is_array, extra_header;

#ifdef VERBOSE_DECODING
      printf ("    %#x: ", hunk);
#endif

      // Extract the type id and other header flags
      type_id = hunk->get_type_id();
      is_array = hunk->get_array_flag();
      extra_header = hunk->get_extra_header_flag();
      obj_size = hunk->size;

      // Using the type id, get the pointer to the fast type
      // descriptor for this object
      FastTDesc *td = ps->fast_tdescs->get_element(type_id)->td;

#ifdef VERBOSE_DECODING
      printf ("%u <%s>", obj_size,ps->fast_tdescs->get_element(type_id)->name);
      printf (" : [%u bytes (%u ptrs)] + [%u bytes (%u ptrs)]\n",
	      td->fixed.size_of_area_of_interest,
	      td->fixed.num_interesting_fields,
	      td->variable.size_of_area_of_interest,
	      td->variable.num_interesting_fields);
#endif

      // Interpret the type descriptor with its flags. If the array
      // flag is set, the type descriptor represents one element of
      // the array. Otherwise, the type descriptor represents a single
      // object.
      if (is_array)
	decode_fast_tdesc_as_array (obj, obj_size, td, extra_header, dir);
      else
      {
	// Flag `extra_header' implies `is_array'.
	assert (!extra_header);
	
	decode_fast_tdesc (obj, obj_size, td, dir);
      }
    }
  }
}



////////////////////////////////////////////////////////////////
// Swizzler::decode_fast_tdesc
//
// Decode the specified object, using its fast type descriptor. This
// version works only for small objects (i.e. objects not larger than
// a page).
//

void Swizzler::decode_fast_tdesc (VirtAddress obj, unsigned long obj_size,
				  FastTDesc *td, SwizzleDirection dir)
{
  // Short circuit this whole thing if there are no pointers in either
  // the fixed or variable parts.
  if (td->fixed.num_interesting_fields == 0
      && td->variable.num_interesting_fields == 0)
    return;

  VirtAddress ptr = obj;
  VirtAddress end_of_obj = obj + obj_size;

  // First decode the fixed part completely.
  decode_offsets_wholly_contained (&td->fixed, ptr, dir);

  // If there exists a variable part, and while the end of the object
  // is not reached, use the type information in the variable part to
  // decode the rest of the object.
  if (td->variable.size_of_area_of_interest > 0)
  {
    // Set the pointer to skip past the fixed part size. We are now
    // pointing at the beginning of the variable part.
    ptr = obj + td->fixed.size_of_area_of_interest;
  
    while ((ptr + td->variable.size_of_area_of_interest) <= end_of_obj)
    {
      // Decode the variable part (represents one element of the
      // array).
      decode_offsets_wholly_contained (&td->variable, ptr, dir);

      // Increment the pointer to skip over the part just decoded.
      ptr = ptr + td->variable.size_of_area_of_interest;
    }
  }
}



////////////////////////////////////////////////////////////////
// Swizzler::decode_fast_tdesc_as_array
//
// Decode a type descriptor interpreted as an array element. This is
// used for dynamically allocated arrays, where the type descriptor at
// the top of the array object indicates the type information for one
// element of the array. In such cases, if there is a variable part in
// the type descriptor, it is really part of the array element,
// because the outermost structure is the whole array itself.
//
// Variable `extra_header' is TRUE iff an extra EXTRA_HEADER_SIZE-byte
// header was prepended to the object by the compiler.
// 

void Swizzler::decode_fast_tdesc_as_array (VirtAddress obj, 
					   unsigned long obj_size, 
					   FastTDesc *td, Boolean extra_header,
					   SwizzleDirection dir)
{
  // Short circuit this whole thing if there are no pointers in either
  // the fixed or variable parts.
  if (td->fixed.num_interesting_fields == 0
      && td->variable.num_interesting_fields == 0)
    return;

  VirtAddress ptr = obj;

  // Adjusting for the extra header (if it exists) is straightforward;
  // for our purposes, we simply ignore it (but it's size *is*
  // included in the object size, so we have to reduce that
  // accordingly).
  if (extra_header)
  {
    // Skip over the extra header
    ptr = ptr + EXTRA_HEADER_SIZE;

    // Reduce total object size accordingly
    obj_size = obj_size - EXTRA_HEADER_SIZE;
  }

  // Calculate the end of the object (array).
  VirtAddress end_of_obj = ptr + obj_size;

  // Calculate the size of one element from the type descriptor.
  size_t element_size = td->fixed.size_of_area_of_interest +
    td->variable_count * td->variable.size_of_area_of_interest;

  // Repeatedly interpret the type descriptor until we are done with
  // the whole array. We can do this easily by calling
  // "decode_fast_tdesc" which decodes one small object with the
  // specified type descriptor.
  while (ptr < end_of_obj)
  {
    // Sanity check -- we have to be inside the object boundary all
    // the time.
    assert (ptr + element_size <= end_of_obj);

    // Decode one object in the array.
    decode_fast_tdesc (ptr, element_size, td, dir);

    // Skip over the object just decoded, continue to the next one.
    ptr = ptr + element_size;
  }
}



////////////////////////////////////////////////////////////////
// Swizzler::decode_large_object_page
//
// Decode one page of a large object (object that spans more than one
// page). If this is not the first page of the object, then it is
// assumed that the first page (the page at the virtual memory
// equivalent to start_pa) is already loaded, and hence the type
// descriptor is available from it.
//

void Swizzler::decode_large_object_page (VirtAddress va, PersAddress pa,
					 PersAddress start_pa,
					 SwizzleDirection dir)
{
  assert (va == (pvmap->get_virt_address (PAGE_ALIGN (pa)) +
		 PAGE_OFFSET (pa)));

  // Find the virtual address corresponding to the start of the
  // object. This can also be done by looking up the virtual address
  // in the VtoPMap.
  VirtAddress start_va = va - (pa - start_pa);
  assert (start_va == (pvmap->get_virt_address (PAGE_ALIGN (start_pa)) +
		       PAGE_OFFSET (start_pa)));

  MemBlockHdr *hunk;

#ifdef VERBOSE_DECODING
  printf ("Decoding large obj page: va %#x, pa %#x (based at %#x), dir %s\n",
	  va, pa, start_pa, (dir == VA2PA) ? "v->p" : "p->v");
#endif

  hunk = (MemBlockHdr *) start_va;
  if (hunk->get_alloc_flag() == ISFREE)
  {
    // This chunk is free, and so the only pointer is the "next"
    // field in the header used for stringing free chunks together. We
    // need to decode that only if this is the first page of the large
    // object.
    if (pa == start_pa)
      convert_pointer (dir, (VirtAddress) &hunk->next);
  }
  else
  {
    // The chunk is allocated and we have to decode the page based on
    // the type descriptor from the first page.
    assert (hunk->get_alloc_flag () == ISALLOC ||
	    hunk->get_alloc_flag () == ISMEMALIGN);

    PStore *ps = pvmap->ps;
    unsigned long obj_size;
    unsigned type_id;
    Boolean is_array, extra_header;

    VirtAddress zone_end, obj_va;
    PersAddress obj_start_pa, obj_pa;

    // We are only going to swizzle one page, so the decoding zone
    // ends at the next page.
    zone_end = va + HEAP_PAGE_SIZE;

#ifdef VERBOSE_DECODING
    printf ("    %#x: ", hunk);
#endif

    // Extract the type id and other header flags.
    type_id = hunk->get_type_id();
    is_array = hunk->get_array_flag();
    extra_header = hunk->get_extra_header_flag();
    obj_size = hunk->size;

    // Using the type id, get the pointer to the fast type
    // descriptor for this object
    FastTDesc *td = ps->fast_tdescs->get_element(type_id)->td;

#ifdef VERBOSE_DECODING
    printf ("%u <%s>", obj_size, ps->fast_tdescs->get_element(type_id)->name);
    printf (" : [%u bytes (%u ptrs)] + [%u bytes (%u ptrs)]\n",
	    td->fixed.size_of_area_of_interest,
	    td->fixed.num_interesting_fields,
	    td->variable.size_of_area_of_interest,
	    td->variable.num_interesting_fields);
#endif

    // Calculate how many total header bytes we have to skip
    unsigned skip_header_bytes = sizeof (MemBlockHdr);
    if (extra_header)
      skip_header_bytes = skip_header_bytes + EXTRA_HEADER_SIZE;

    // This variable holds the pointer to the "real" beginning of the
    // object past any headers (our own malloc header and any other
    // invisible headers added by the compiler).
    obj_start_pa = start_pa + skip_header_bytes;
    
    // Set obj_pa and obj_va to positions corresponding to where we
    // have to start decoding. If we are on the first page, this is
    // just past the header (and any additional extra "invisble"
    // headers added by the compiler). Otherwise, we are somewhere in
    // the middle of the object and we start decoding at the address
    // that was passed to this function.
    if (pa == start_pa)
    {
      // First page of the large object.
      obj_pa = obj_start_pa;
      obj_va = va + skip_header_bytes;
    }
    else
    {
      // Not first page of large object.
      obj_pa = pa;
      obj_va = va;
    }

    // If "is_array" flag is set, this type descriptor represents one
    // a large array of objects. Otherwise, it's a single large object.
    if (is_array)
      decode_large_array_object (obj_va, obj_pa, obj_start_pa, 
				 zone_end, dir, td); 
    else
    {
      // Flag `extra_header' implies `is_array'.
      assert (!extra_header);

      decode_large_object (obj_va, obj_pa, obj_start_pa, zone_end, dir, td); 
    }
  }
}



////////////////////////////////////////////////////////////////
// Swizzler::decode_large_object
//
// Decode a single large object. By definition the object spans
// multiple pages, but the object itself is not an array of large
// objects. The first two arguments specify the virtual and persistent
// addresses of a large object page to be decoded. This page may be in
// the middle of the object. The third argument is a pointer to the
// start of the large object (i.e. first page of the large object).
// The fourth argument is the end of the current decoding region. Note
// that the first argument is essentially the beginning of the current
// decoding region.
//

void Swizzler::decode_large_object (VirtAddress va, PersAddress pa,
				    PersAddress start_pa, VirtAddress limit_va,
				    SwizzleDirection dir, FastTDesc *td)
{
  assert (va == (pvmap->get_virt_address (PAGE_ALIGN (pa)) +
		 PAGE_OFFSET (pa)));

  // Calculate the persistent address of the end of the object.
  // Another method of doing this is to lookup the VtoPMap, but
  // addition is faster than hash table lookup. If we are at the last
  // page of the large object, then limit_va would correspond to the
  // address of the page after the object, in which case, the lookup
  // for limit_pa will return NOT_A_MAPPING.
  PersAddress limit_pa = pa + (limit_va - va);
#ifndef NDEBUG
  PersAddress tmp_pa
    = v_to_p_map->get_pers_address ((VirtAddress) PAGE_ALIGN (limit_va));
  
  assert (tmp_pa == (PersAddress) NOT_A_MAPPING || 
	  limit_pa == tmp_pa + PAGE_OFFSET (limit_va));
#endif

#ifdef VERBOSE_DECODING
  printf ("decode (%c): va = [%#x - %#x] pa = [%#x - %#x] based at %#x\n", 
	  dir, va, limit_va, pa, limit_pa, start_pa);
  printf ("    td: fixed (size %#x + [%d:]) var %d * (size %#x + [%d:])\n",
	  td->fixed.size_of_area_of_interest,
	  td->fixed.num_interesting_fields,
	  td->variable_count, td->variable.size_of_area_of_interest,
	  td->variable.num_interesting_fields);
#endif

  // Four major cases are possible, based on where the region to be
  // decoded overlaps with the large object. See figures in file
  // "largeobj.ps" in "docs" directory which illustrate the different
  // cases graphically.

  if ((pa == start_pa) && 
      ((limit_pa - start_pa) >= td->fixed.size_of_area_of_interest))
  {
    // Case I:
    // The start of the region coincides with the start of the large
    // object, fully covers the fixed part of the large object, and
    // partially covers zero or more elements in the variable part.

    // Decode all of the fixed part, since we know it is completely
    // covered by the region to be decoded.
    decode_offsets_wholly_contained (&td->fixed, va, dir);

    // If the object has variable part, decode that also, taking care
    // that the end of the decoding region might only partially
    // overlap an element.
    if (td->variable.num_interesting_fields != 0)
    {
      // Make sure that the elements in the variable part are not zero
      // sized.
      assert (td->variable.size_of_area_of_interest != 0);
    
      // Skip over the fixed part, and initialize a pointer to the
      // beginning of the variable part.
      VirtAddress ptr = va + td->fixed.size_of_area_of_interest;
    
      // The variable part of the type descriptor specifies the type
      // information for a single element in the variable part. Loop
      // until all the fully included elements are decoded. Each such
      // element has all of its offsets contained in the decoding
      // region.
      while ((ptr + td->variable.size_of_area_of_interest) < limit_va)
      {
	decode_offsets_wholly_contained (&td->variable, ptr, dir);
	ptr += td->variable.size_of_area_of_interest;
      }

      // Handle partial element, if any, at the end of the decoding
      // region. Such an element only has some of its initial offsets
      // in the decoding region.
      decode_offsets_initial_contained (&td->variable, ptr, limit_va, dir);
    }
  }
  else
  {
    if ((pa - start_pa) > td->fixed.size_of_area_of_interest)
    {
      // Case II:
      // The decoding region does not overlap the fixed part of the
      // object at all. This means that it is entirely contained
      // within the variable part. Two subcases are possible,
      // depending on whether the region spans one or more element
      // boundaries in the variable part, or is entirely contained in
      // one element.

      // If the object has no interesting fields in the variable part,
      // simply return, because no swizzling is necessary.
      if (td->variable.num_interesting_fields == 0)
	return;

      // Make sure that the elements in the variable part are not zero
      // sized.
      assert (td->variable.size_of_area_of_interest != 0);
    
      // Since the decoding region could start in the middle of an
      // element in the variable part, we need to find the
      // corresponding beginning of the element. We calculate this by
      // calculating offset of the beginning of the region in the
      // element, and subracting the offset from the persistent and
      // virtual addresses of the begin point, to get the beginning of
      // the element.
      unsigned long offset;
      offset = (pa - start_pa - td->fixed.size_of_area_of_interest) % 
	td->variable.size_of_area_of_interest;

      // Calculate the persistent and virtual address for the
      // beginning of the element in the variable part.
      PersAddress elem_pa;
      VirtAddress elem_va;
      elem_pa = pa - offset;
      elem_va = va - offset;		
    
      // Check whether the region spans at least one element boundary
      // or not. Since we know the start of the element and the size,
      // we can compare against the end of the decoding region to
      // decide which subcase is true.
      if (limit_va >= (elem_va + td->variable.size_of_area_of_interest))
      {
	// Subcase a: 
	// The region spans at least one element boundary in the
	// variable part. This means, we have to first decode the
	// final offsets of the current element (the region possibly
	// starts in the middle of the element), then decode zero or
	// more elements fully contained in the region, and then zero
	// or one element whose initial offsets are contained in the
	// region.

	// Decode final contained offsets of the first element.
	decode_offsets_final_contained (&td->variable, elem_va, va, dir);
      
	// Skip over the element just decoded.
	elem_va += td->variable.size_of_area_of_interest;
      
	// Loop through zero or more elements of the variable part
	// that are fully contained in the region. Each such element
	// has all of its offsets contained in the decoding region.
	while ((elem_va + td->variable.size_of_area_of_interest) < limit_va)
	{
	  decode_offsets_wholly_contained (&td->variable, elem_va, dir);
	  elem_va = elem_va + td->variable.size_of_area_of_interest;
	}

	// Decode initial contained offsets of the last element, if
	// any. Such an element only has some of its initial offsets
	// in the decoding region.
	decode_offsets_initial_contained (&td->variable, elem_va, 
					  limit_va, dir);
      }
      else
      {
	// Subcase b: 
	// The region is entirely within one element of the variable
	// part. We only have to decode the middle offsets that are
	// contained in the region.
      
	decode_offsets_middle_contained (&td->variable, elem_va, va, 
					 limit_va, dir);
      }
    }
    else
    {
      if ((limit_pa - start_pa) >= td->fixed.size_of_area_of_interest)
      {
	// Case III:
	// The decoding region overlaps with some of the fixed part of
	// the object (however, *not* from the beginning, since that
	// situation was caught by Case I), and some of the variable
	// part. Thus, we still have some of the fixed part to decode
	// before we can start decoding repetitions of the variable
	// part. We ARE sure that the fixed/variable boundary is in
	// the decoding region. This is similar to Case I except that
	// the decoding region doesn't start at the beginning of the
	// object.
  
	// Find the virtual address corresponding to the start of the
	// object. This can also be done by looking up the virtual
	// address in the VtoPMap.
	VirtAddress start_va = va - (pa - start_pa);
	assert (start_va == (pvmap->get_virt_address (PAGE_ALIGN (start_pa)) +
			     PAGE_OFFSET (start_pa)));

	// First, decode the final contained offsets of the fixed
	// part.
	decode_offsets_final_contained (&td->fixed, start_va, va, dir);

	// If the object has variable part, decode that also, taking
	// care that the end of the decoding region might only
	// partially overlap an element.
	if (td->variable.num_interesting_fields != 0)
	{
	  // Make sure that the elements in the variable part are not
	  // zero sized.
	  assert (td->variable.size_of_area_of_interest != 0);
    
	  // Skip over the fixed part, and initialize a pointer to the
	  // beginning of the variable part.
	  VirtAddress ptr = start_va + td->fixed.size_of_area_of_interest;
  
	  // Loop through zero or more elements of the variable part
	  // that are fully contained in the region. Each such element
	  // has all of its offsets contained in the decoding region.
	  while ((ptr + td->variable.size_of_area_of_interest) <= limit_va)
	  {
	    decode_offsets_wholly_contained (&td->variable, ptr, dir);
	    ptr = ptr + td->variable.size_of_area_of_interest;
	  }
    
	  // Decode initial contained offsets of the last element, if
	  // any. Such an element only has some of its initial offsets
	  // in the decoding region.
	  decode_offsets_initial_contained (&td->variable, ptr, limit_va, dir);
	}
      }
      else
      {
	// Case IV:
	// The decoding region is entirely within the fixed part of
	// the object. We only have to decode those middle offsets of
	// the fixed part which are contained in the region. However,
	// the decoding region could start from the beginning of the
	// fixed part. We are only sure that it doesn't span the
	// boundary between fixed and variables parts. In any case,
	// attempting to decode middle offsets will take care of the
	// case where the region starts from the beginning.

	// Find the virtual address corresponding to the start of the
	// object. This can also be done by looking up the virtual
	// address in the VtoPMap.
	VirtAddress start_va = va - (pa - start_pa);
	assert (start_va == (pvmap->get_virt_address (PAGE_ALIGN (start_pa)) +
			     PAGE_OFFSET (start_pa)));
	// Decode the middle offsets of the fixed part.
	decode_offsets_middle_contained (&td->fixed, start_va, va, 
					 limit_va, dir);
      }
    }
  }
}



////////////////////////////////////////////////////////////////
// Swizzler::decode_large_array_object
//
// Decode a large object which is an array of objects i.e. flag
// "is_array" is set in the header of this object. In this case, the
// type descriptor passed to this function represents type
// information for only one element of the array.
//

void Swizzler::decode_large_array_object (VirtAddress va, PersAddress pa,
					  PersAddress start_pa,
					  VirtAddress limit_va, 
					  SwizzleDirection dir, FastTDesc *td)
{
  assert (va == (pvmap->get_virt_address (PAGE_ALIGN (pa)) +
		 PAGE_OFFSET (pa)));

  // Calculate the persistent address of the end of the object.
  // Another method of doing this is to lookup the VtoPMap, but
  // addition is faster than hash table lookup. If we are at the last
  // page of the large object, then limit_va would correspond to the
  // address of the page after the object, in which case, the lookup
  // for limit_pa will return NOT_A_MAPPING.
  PersAddress limit_pa = pa + (limit_va - va);

#if 0
// Probably a mistaken assertion, according to Sheetal on 2/13/95.
// (Steve Rozen)
#ifndef NDEBUG
  PersAddress tmp_pa
    = v_to_p_map->get_pers_address ((VirtAddress) PAGE_ALIGN (limit_va));
  
  assert (tmp_pa == (PersAddress) NOT_A_MAPPING || 
	  limit_pa == tmp_pa + PAGE_OFFSET (limit_va));
//
#endif
#endif

  // Calculate the size of the element from the type descriptor.
  size_t element_size = td->fixed.size_of_area_of_interest + 
    td->variable_count * td->variable.size_of_area_of_interest;

  // The elements cannot be zero sized.
  assert (element_size != 0);
  
#ifdef VERBOSE_DECODING
  printf ("(elem size %#x) ", element_size);
#endif
  
  // There are two cases to be distinguished. Either the region to be
  // decoded lies completely within one element of the array, or the
  // region to be decoded spans at least one element boundary. See
  // figure in file "largeobj.ps" in "docs" directory for a graphical
  // representation.

  unsigned long skip = (pa - start_pa) % element_size;
  
  // The variable current_elem_limit holds the virtual address of the
  // end of the array element that pa (and va) point into.
  VirtAddress current_elem_limit = va - skip + element_size;
  
  if (limit_va <= current_elem_limit)
  {
    // Case II: 
    // The region to be decoded lies entirely within a single array
    // element. This case is handled by deferring to the other
    // decoder, but modifying the pointers so it thinks that this is
    // just a single object rather than an array element; this is an
    // advantage to NOT having that decoder read the object header.

    decode_large_object (va, pa, pa - skip, limit_va, dir, td);	
  }
  else
  {
    PersAddress elem_pa;
    VirtAddress elem_va;
    
    // Case I: 
    // The region to be decoded spans at least one array element
    // boundary. This case is handled by deferring to the other
    // handler for the endpoints (partial elements at each end of the
    // decoding region), but otherwise just looping over the fields of
    // interest for the wholly contained array elements. Variable
    // "elem_pa" is the persistent address of the beginning of the
    // current element, and "elem_va" is the corresponding virtual
    // address.
    elem_pa = pa - skip;
    elem_va = va - skip;
    
    // First portion -- partial element at the beginning of the region.

    decode_large_object (va, pa, elem_pa, current_elem_limit, dir, td);

    // Skip over the (partial) element just decoded.
    elem_pa += element_size;
    elem_va += element_size;
    
    // Middle portion. These may contain zero or more elements that we
    // have to fully decode. Iterate as long as the next
    // "element_size" part is wholly contained within the decoding
    // region.
    while ((elem_va + element_size) <= limit_va)
    {
#ifdef VERBOSE_DECODING
      printf ("Middle portion: va %#x, pa %#x\n", elem_va, elem_pa);
#endif
      // This is the type descriptor interpreted as an array element,
      // so it is a fixed part plus 'n' variable parts.
      
      // Decode the fixed part completely.
      decode_offsets_wholly_contained (&td->fixed, elem_va, dir);

      // Skip over the fixed part, and initialize a pointer to the
      // beginning of the variable part.
      VirtAddress ptr = elem_va + td->fixed.size_of_area_of_interest;

      for (unsigned int j = 0; j < td->variable_count; j++)
      {
	decode_offsets_wholly_contained (&td->variable, ptr, dir);
	ptr = ptr + td->variable.size_of_area_of_interest;
      }
      
      // Go on to next element in the array.
      elem_pa += element_size;
      elem_va += element_size;
    }
    
    // Last portion -- partial element at the end of the decoding
    // region.
#ifdef VERBOSE_DECODING
    printf ("Last portion: va %#x, pa %#x\n", elem_va, elem_pa);
#endif
#if 0
    decode_large_object (elem_va, elem_pa, elem_pa, limit_va, dir, td);
#endif
// The following code is suggested by Sheetal on Feb. 13, 1995:
    if (elem_va < limit_va)
      decode_large_object (elem_va, elem_pa, elem_pa, limit_va, dir, td);

  }
}
