/* 
 * Copyright (c) 1995
 *       Whitehead Institute for Biomedical Research. All rights reserved.
 *
 * Released under terms specified in the Copyright file present in the top
 * directory of this distribution.
 */


#include <memory.h>

#include "list.H"

#ifndef BASICS_H
#include "basics.H"
#endif

static const int initial_size = 8;

Generic_List::Generic_List() :
_count (0), current_size(initial_size)
{
    _data = PNEW_ARRAY_ADDR(this,void*,current_size);
    ABORT ((_data == NULL), "Could not allocate space to construct Array");
    memset(_data, 0, current_size * sizeof(*_data));
    return ;
}

Generic_List::Generic_List(int size) :
_count(0), current_size(size)
{
    _data = PNEW_ARRAY_ADDR(this,void*,current_size);
    ABORT ((_data == NULL), "Could not allocate space to construct Array");
    memset(_data, 0, current_size * sizeof(*_data));
    return ;
}

int Generic_List::index(const void* element) const 
{
    for (int i = low (); i <= high (); i++)
      {
	  if (get (i) == element) return i;
      }
    // If fall out of loop, element not in array
    //
    return low () - 1;
}

void* Generic_List::remove(int index)
{
  ABORT ((index < low()),
	 "Array index " << index <<
	 " too small (<" << low() <<
	 ") in Generic_List::remove(int)");
  ABORT ((index > high ()),
	 "Array index " << index <<
	 " too big (>" << high() <<
	 ") in Generic_List::remove(int)");
    
  void* result =  _data [index];

  for (int dst_index = index, src_index = index + 1;
       src_index <= high ();
       dst_index++, src_index++) {
    _data [dst_index] = _data [src_index];
  }
  _count --;
  return result;
}

void Generic_List::put(int index, const void* element)
{
  // There are three cases:
  //   1.  index < next_free_idx
  //       This is the normal case.
  //       Slide elements up to make room for new guy.
  //   2.  index == next_free_idx
  //       New element goes at the end.
  //       This is handled by the same code as case 1.
  //   3.  index > next_free_idx
  //       Fill with NULLs up to index, then put new element.
  
  ABORT ((index < 0),
	 "List index too small (< 0) in Generic_List::put (int, void*)");
  
  int high_index = high ();
  if (index <= high_index)
    {
      for (int i = high_index; i >= index; i--)
	{
	  put_replace(i+1, get(i));
	}
      }
  else
    {
      for (int i = high_index + 1; i < index; i++)
	{
	  put_replace(i, NULL);
	}
    }
  put_replace(index, element);
  return ;
}


void Generic_List::put_replace(int index, const void* element)
{
   if (index >= current_size) put_expand (index);
    // It's safe to assume there is space for the new element.
    _data[index] = (void*) element;
    _count = max (index+1, _count);
}


void* Generic_List::remove(const void* element)
{
  int i = index(element);
  return i >= low() ? remove(i) : NULL;
}


void *Generic_Cursor::first()
{
  _cursor_index = 0;
  void *r = A->count() > 0 ? A->_data[0] : NULL;
  _cursor_index++;
  return r;
}


void *Generic_Cursor::next()
{
  void *r = (_cursor_index < A->count()) ?A->_data[_cursor_index] : NULL;
  _cursor_index++;
  return r;
}


void *Generic_Cursor::last()
{
  _cursor_index = A->count() - 1;
  void *r = (_cursor_index >= 0) ?A->_data[_cursor_index] : NULL;
  _cursor_index--;
  _last_called = 1;
  return r;
}


void *Generic_Cursor::previous()
{
  ABORT(!_last_called, "Generic_Cursor::last() not called ");
  void *r = (_cursor_index >= 0) ?A->_data[_cursor_index] : NULL;
  _cursor_index--;
  return r;
}


Generic_Cursor *Generic_List::cursor_of() const 
{
  return new Generic_Cursor((const Generic_List *) this);
}


void Generic_List::put_expand(int index)
{
    // The index lies outside the currently allocated space.
    // Allocate more space.
    // Copy the old space to the new space.
    // Finally, delete the old space, and go on.

    do {
      current_size = 2 * current_size;
    } while (index >= current_size);
    
    void**  old_data = _data;
    
    _data = PNEW_ARRAY_ADDR(this,void*,current_size);
    memset(_data,0,sizeof(void*) * current_size);
    ABORT((_data == NULL),
	  "Could not allocate space in Generic_List::put_expand(int)");
      
    for (int i = 0; i < count(); i++) _data[i] = old_data[i];
    delete old_data;
}


