/* 
 * Copyright (c) 1995
 *       Whitehead Institute for Biomedical Research. All rights reserved.
 *
 * Released under terms specified in the Copyright file present in the top
 * directory of this distribution.
 */


// $Id: material_set.C,v 1.5 1995/02/13 17:50:02 steve Exp $
#ifndef SAFE_H
#include "safe.H"
#endif

#ifndef BASICS_H
#include "basics.H"
#endif

#ifndef MATERIAL_SET_H
#include "material_set.H"
#endif

#ifndef QL_TERM_H
#include "ql_term.H"
#endif

#ifndef QL_ARGS_H
#include "ql_args.H"
#endif

#ifndef MATERIAL_H
#include "material.H"
#endif

#ifndef VALUE_H
#include "value.H"
#endif

#ifndef BUILTIN_H
#include "builtin.H"
#endif

// class Material_Set : public List_Material
// {};

class Set_Value : public List_Value
{
public:
  int set_put(Value* element);
  // Return 1 on success and 0 on error.  Use errmsg() to get the error text
  // after an error.  The definition of this function constrains value sets to
  // contain only elements that are mutually comparable by ==.

  int set_remove(const Value* element);
  // Return 1 on success and 0 on error.  Use errmsg() to get the error text
  // after an error. Equality is based on the the == operator for values.

};

static int
is_copied(const Value *v) {
  return (v->tv_type() != Value::MATERIAL_POINTER &&
	  v->tv_type() != Value::STEP_POINTER);
}

int
Set_Value::set_remove(const Value* v)
{
  clear_error();
  for(int i=low(); i <= high(); i++) {
    int equal = (*v == *get(i));
    if (errmsg()) return 0;
    if (equal) { 
      if (is_copied(get(i))) get(i)->unsave();
      remove(i);
      return 1;
    }
  }
  return 1;
}

int
Set_Value::set_put(Value* v)
{
  clear_error();
  for(int i=low(); i <= high(); i++) {
    int equal = (*v == *get(i));
    if (errmsg()) return 0;
    if (equal) return 1;
  }
  if (is_copied(v)) {
    P_SEGMENT *Seg = 
#if defined(OBJECTSTORE)
    os_segment::of(this)
#elif defined(TEXAS)
    obj_store_of(this)
#else
    NULL
#endif
    ;
    put(v->save(Seg));
  } else put(v);
  return 1;
}


class All_Sets_Entry {
public:
  All_Sets_Entry(const String *name);
  ~All_Sets_Entry();
  SLOT(char*,name);
  SLOT(Set_Value*,set);
  // The following might be needed in the future.
  SLOT(QL_Type_Descriptor*,type_descriptor);
};


All_Sets_Entry::All_Sets_Entry(const String *name)
: _set(PNEW_ADDR(this,Set_Value,)), _type_descriptor(0)
{
  char *n = name->get();
  _name = PNEW_ARRAY_ADDR(this,char,strlen(n) + 1);
  safe_strcpy(_name,n);
  delete n;
}


All_Sets_Entry::~All_Sets_Entry()
{
  delete [] name();
  delete set();
}


static Value_Set_Table *temp_v_sets = new Value_Set_Table;


static All_Sets_Entry *
find_entry(const String *name, Value_Set_Table *all_sets)
{
  if (!all_sets) return 0;
  Cursor_All_Sets_Entry *c = all_sets->cursor_of();
  char *n = name->get();

  for (All_Sets_Entry *e = c->first(); e; e = c->next()) {
    if (!strcmp(e->name(),n)) { delete n; delete c; return e; }
  }
  delete n; delete c; return 0;
}


static Set_Value *
find_value_set(const String *name, Value_Set_Table *all_sets)
{
  All_Sets_Entry *e = find_entry(name, all_sets);
  return e ? e->set() : 0;
}


static Set_Value *
find_or_create_Set_Value(const String *name, Value_Set_Table *all_sets)
{
  Set_Value *r = find_value_set(name, all_sets);
  if (r) return r;
  
  All_Sets_Entry *e  = PNEW_ADDR(all_sets,All_Sets_Entry,(name));
  all_sets->put(e);
  return e->set();
}


class Q_State
{
 public:
  Value **_m;
  int _size;
  int _i;
  Q_State(Set_Value *s)
    : _size(s->count()), _i(-1) 
      {
	_m = new Value * [ _size ];
	Cursor_Value *c = s->cursor_of();
        int j = 0;
	for (Value *m = c->first(); m; j++, m = c->next()) _m[j] = m;
	delete c;
      }
  ~Q_State() { delete [] _m; }
  Value *next() { _i++; if (_i >= _size) return 0; else return _m[_i]; }
};


static QL_Iterator_State *
Set_Value_bf_pred_first(QL_Term* t, int m, Value*in[], int n, Value*out[],
			   Value_Set_Table *all_sets)
{
  abort_m_n(m, 1,n,1);
  ABORT(!in[0],"0 illegal here");
  if (!check_arg_type(t,&in[0],Value::STRING,0)) return 0;
  const String *name = in[0]->down_to_STRING();
  Set_Value *s = find_value_set(name,all_sets);
  if (!s) return 0;
  Q_State *state = new Q_State(s);
  Value *v = state->next();
  if (!v) { delete state; return 0; }
  out[0] = v;
  return (QL_Iterator_State*) state;
}


QL_FIRST_HEADER(v_set_bf_pred_first)
{
  return Set_Value_bf_pred_first(t, m, in, n, out, QL_DD::persistent_m_sets());
}


QL_FIRST_HEADER(temp_v_set_bf_pred_first)
{
  return Set_Value_bf_pred_first(t, m, in, n, out, temp_v_sets);
}


QL_NEXT_HEADER(v_set_pred_next)
{
  Q_State *state = (Q_State*) s;
  Value *v = state->next();
  if (!v) { delete state; return 0; }
  out[0] = v;
  if (in || m || n || t); // Suppress warnings.
  return s;
}


static inline String* insert_delete_args(QL_Term *t)
{
  QL_Args *args = t->args();
  Value *tv = args->value();
  if (!tv) { ql_eval_error(t, "First argument must be bound for"); 
	     return 0; }
  if (tv->tv_type() != Value::STRING) {
    ql_eval_error(t, "First argument must be a 'STRING' for");
    return 0;
  }
  return tv->down_to_STRING();
}


static Value *
extract_value(QL_Term* t)
{
  QL_Args *material_arg = t->args()->next();
  Value *v = material_arg->value();
  if (!v) { ql_eval_error(t, "Second argument must be bound for"); return 0; }
  return v;
}


QL_Iterator_State *value_set_insert(QL_Term*t, Value_Set_Table *all_sets)
{
  if (t->args()->arity() != 2) {
    ql_eval_error(t, "Expected a value set name (a 'STRING') and a value as arguments for");
    return 0;
  }
  String *name = insert_delete_args(t);
  if (!name) return 0;
  Set_Value *s = find_or_create_Set_Value(name, all_sets);
  Value *v = extract_value(t);
  if (!v) return 0;
  if (!s->set_put(v)) { ql_eval_error(t, errmsg(), " in"); return 0; }
  return (QL_Iterator_State*)1;
}



QL_Iterator_State *v_set_insert(QL_Term*t)
{
  return value_set_insert(t, QL_DD::persistent_m_sets());
}


QL_Iterator_State *temp_v_set_insert(QL_Term*t)
{
  return value_set_insert(t, temp_v_sets);
}


// Delete an element from a value set.
QL_Iterator_State *value_set_delete(QL_Term*t, Value_Set_Table* all_sets)
{
  int arity = t->args()->arity();
  if (arity != 2) {
    ql_eval_error(t, "Expected a value set name (a 'STRING') and a value as arguments for");
    return 0;
  }
  const String *name = insert_delete_args(t);

  if (!all_sets) return 0;
  // We put this after insert_delete_args so that error messages will
  // be the same whether or not all_sets has been initialized.

  if (!name) return 0;
  QL_Args *material_arg = t->args()->next();
  All_Sets_Entry *e = find_entry(name,all_sets);
  if (e){
    Value *v = extract_value(t);
    if (!v) return 0;
    e->set()->set_remove(v);
  }
  return (QL_Iterator_State*)1;
}


QL_Iterator_State *v_set_delete(QL_Term*t)
{
  return value_set_delete(t, QL_DD::persistent_m_sets());
}


QL_Iterator_State *temp_v_set_delete(QL_Term*t)
{
  return value_set_delete(t, temp_v_sets);
}


static QL_Iterator_State* 
delete_entire(const Value *set_name, Value_Set_Table *s)
{
  All_Sets_Entry *e = find_entry(set_name->down_to_STRING(),s);
  if (e) {
    // Careful!!
    // We need to delete non-Material, non-Step values,
    // since these have to be shared.
    Cursor_Value *c = e->set()->cursor_of();
    for (Value *v = c->first(); v; v = c->next())
      if (is_copied(v)) v->unsave();
    delete c;
    delete e;
    s->remove(e);
  }
  return (QL_Iterator_State *)1; // delete_entire{,_temp}_Set_Value is
				 // always true.
}


static Value::tv_type_id one_string[] = { Value::STRING };


BUILTIN_FIRST_DEFINITION(delete_entire_value_set_pred,1,0,one_string,1,1)
{
  if (t || m || n || out_tv); // Suppress warnings.
  return delete_entire(in_tv[0], QL_DD::persistent_m_sets());
}


BUILTIN_FIRST_DEFINITION(delete_entire_temp_value_set_pred,
			 1,0,one_string,1,1)
{
  if (t || m || n || out_tv); // Suppress warnings.
  return delete_entire(in_tv[0], temp_v_sets);
}


static QL_Iterator_State *
value_set_bb_pred_first(QL_Term *t, int m, Value *in[], int n, Value *[],
			   Value_Set_Table *all_sets)
{
  abort_m_n(m, 2, n, 0);
  if (!check_arg_type(t,&in[0],Value::STRING,0)) return 0;
  if (!check_arg_type(t,&in[1],Value::MATERIAL_POINTER,1)) return 0;
  const String *name = in[0]->down_to_STRING();
  Set_Value *s = find_value_set(name,all_sets);
  if (!s) return 0;
  if (s->contains(in[1]->down_to_MATERIAL_POINTER()))
    return (QL_Iterator_State*) 1;
  else return 0;
}


QL_FIRST_HEADER(v_set_bb_pred_first)
{
  return value_set_bb_pred_first(t, m, in, n, out, QL_DD::persistent_m_sets());
}


QL_FIRST_HEADER(temp_v_set_bb_pred_first)
{
  return value_set_bb_pred_first(t, m, in, n, out, temp_v_sets);
}


QL_Iterator_State *material_set_f_pred_first(const Value *out[],
					     const Value_Set_Table *all_sets)
{
  Cursor_All_Sets_Entry *c = all_sets->cursor_of();
  All_Sets_Entry *e = c->first();
  if (!e) { delete c; return 0; }
  out[0] = String::make(e->name());
  return (QL_Iterator_State*) c;
}


QL_FIRST_HEADER(v_set_f_pred_first)
{
  if (t || m || in || n); // Suppress warnings.
  return material_set_f_pred_first(out, QL_DD::persistent_m_sets());
}


QL_FIRST_HEADER(temp_v_set_f_pred_first)
{
  if (t || m || in || n); // Suppress warnings.
  return material_set_f_pred_first(out, temp_v_sets);
}


QL_NEXT_HEADER(v_set_f_pred_next)
{
  if (t || m || in || n); // Suppress warnings.
  Cursor_All_Sets_Entry *c = (Cursor_All_Sets_Entry*)s;
  All_Sets_Entry *e = c->next();
  if (!e) { delete c; return 0; }
  out[0] = String::make(e->name());
  return s;
}


QL_Iterator_State *value_set_b_pred(const Value *in,
				       const Value_Set_Table *all_sets)
{
  Cursor_All_Sets_Entry *c = all_sets->cursor_of();
  char *n = in->down_to_STRING()->get();
  for (All_Sets_Entry *e = c->first(); e; e = c->next())
    if (!strcmp(e->name(),n)) {
      delete n; delete c; return (QL_Iterator_State*)1;
    }
  delete n; delete c; return 0;
}


BUILTIN_FIRST_DEFINITION(v_set_b_pred,1,0,one_string,1,1)
{
  if (t || m || n || out_tv); // Suppress warnings.
  return value_set_b_pred(in_tv[0], QL_DD::persistent_m_sets());
}


BUILTIN_FIRST_DEFINITION(temp_v_set_b_pred,1,0,one_string,1,1)
{
  if (t || m || n || out_tv); // Suppress warnings.
  return value_set_b_pred(in_tv[0], temp_v_sets);
}



