/* 
 * Copyright (c) 1995
 *       Whitehead Institute for Biomedical Research. All rights reserved.
 *
 * Released under terms specified in the Copyright file present in the top
 * directory of this distribution.
 */


%{
#include <string.h>
#include <iostream.h>

#ifndef BASICS_H
#include "basics.H"
#endif

#ifndef VALUE_H
#include "value.H"
#endif

#define YYSTYPE void *

#include "ql.tab.h"

// ******  WARNING  ******
// Be sure to add a break
// or return statement after
// every rule because of the
// following #define!
#define YY_BREAK // To avoid ``statement not reached errors''

#define YY_USER_ACTION { error_position += yyleng; }

// ******  WARNING  ******
// Don't use | in rules. It throws
// off the maintenance of error_position
// by executing multiple YY_USER_ACTIONs.
// This multiple execution of YY_USER_ACTIONs is
// probably due to the redefinition of YY_BREAK, above.

#ifndef QL_ERROR_H
#include "ql_error.H"
#endif

#define DEFINE_MAKE_FUNCTION(TYPE)\
static void *CAT_ID2(make_,TYPE)(char *text)\
{\
  clear_error();\
  TYPE *v = TYPE::make(text);\
  if (errmsg()) {\
    ql_parse_error(errmsg(), " reading token ", text);\
    return 0;\
  } else return v; \
}

DEFINE_MAKE_FUNCTION(String)
DEFINE_MAKE_FUNCTION(Int)
DEFINE_MAKE_FUNCTION(Date)
DEFINE_MAKE_FUNCTION(Float)
#undef MAKE_FUNCTION

static String *partial_string = 0;

static int comp_tok_type(const String *sv)
{
  if (sv->get(1) == '=') return TWO_OP;
  else if (sv->get(1) == '>' && sv->get(0) == '<') return TWO_OP;
  else return TOK_STRING;
}


static int tok_type(const String *sv)
{
  if (!sv) return ERROR;
  int high = sv->high();
  if (high <= 1) {
    switch (sv->get(0)) {
    case '=':
    case '>': 
    case '<': if (high == 0) return TWO_OP;
              else return comp_tok_type(sv);
    case ';': if (high == 0) return TWO_OP;
              else if (sv->get(1) == ';') return TWO_OP;
              else return TOK_STRING;
    case '~': return high == 0 ? TWO_OP : TOK_STRING;
    case 'i': if (high == 0) return TOK_STRING;
              else if (sv->get(1) == 's') return TWO_OP;
    }
  }
  return TOK_STRING;
}

int interactive_input = 0;
int fifo_input = 0;
static int print_prompt_now = 1;
static int add_to_line = 0;

#define YY_INPUT(buf,result,max_size) \
        if ( fifo_input ) \
                { \
                    int c; \
                    result = 0; \
                    for (; result < max_size; result++) { \
                             /* result is the # of chars read. */ \
                             c = getc( yyin ); \
                             buf[result] = (char)c; \
                             if ('\n' == c) { result++; break; } \
                             if (EOF == c) break; \
	            } \
                 } \
	else if ( interactive_input ) \
		{ \
                if (print_prompt_now) { \
                   cout << error_position.line()+add_to_line << "> " << flush; \
                   print_prompt_now = 0; \
                } \
		int c = getc( yyin ); \
		if ('\n' == c) print_prompt_now = 1; \
                add_to_line = 1; \
		result = c == EOF ? 0 : 1; \
		buf[0] = (char) c; \
		} \
	else if ( ((result = fread( buf, 1, max_size, yyin )) == 0) \
		  && ferror( yyin ) ) \
		YY_FATAL_ERROR( "input in flex scanner failed" );

int respond_with_control_A = 0;

%}
ws         [ \t]+

comment    %[^\n\001]*

variable   [A-Z][a-zA-Z_0-9]*
unqstring [a-z][a-zA-Z_0-9]*
op_string [+\-*/\\\^<>=`~:.?@#$&]+

integer    [-+]?[0-9]+

float1     [-+]?[0-9]*(\.[0-9]|[0-9]\.)[0-9]*
float2     [-+]?[0-9]*(\.[0-9]|[0-9]\.|[0-9])[0-9]*e[-+]?[0-9]+

date       [0-9:]+

%x QUOTED_STRING
%x ESCAPED_STRING
%x ESCAPED_STRING1

%%
{ws}\(        return WS_L_PAREN;

{ws}         break;

\n\(         { error_position.nl(); return WS_L_PAREN; }

{comment}    break;

{variable}   { yylval = make_String(yytext);
               return yylval ? VARIABLE : ERROR;
              }

{integer}    { yylval = make_Int(yytext);
               return yylval ? TOK_INTEGER : ERROR;
             }

{float1}     { yylval = make_Float(yytext);
               return yylval ? TOK_FLOAT : ERROR;
             }

{float2}     { yylval = make_Float(yytext);
               return yylval ? TOK_FLOAT : ERROR;
             }

\n           error_position.nl(); break;

\001         respond_with_control_A = 1; error_position.reset(); return DOT;

\.           respond_with_control_A = 0; return DOT;

\,           return COMMA;

{unqstring} { yylval = make_String(yytext);
              int r = tok_type((String*) yylval);
              return r;
            }

{op_string}[-+]([0-9]) { yyless(yyleng - 2);
                         yylval = make_String(yytext);
                         int r = tok_type((String*) yylval);
                         return r;
                       }

{op_string}[-+](\.[0-9]) { yyless(yyleng - 3);
                           yylval = make_String(yytext);
                           const long r = tok_type((String*) yylval);
                           return r;
                         }

{op_string} { yylval = make_String(yytext);
              int r = tok_type((String*) yylval);
              return r;
            }

;           { yylval = make_String(yytext);
              int r = tok_type((String*) yylval);
              return r;
            }

;;          { yylval = make_String(yytext);
              int r = tok_type((String*) yylval);
              return r;
            }


\'\'         { yylval = make_String("");
               return yylval ? TOK_STRING : ERROR;
               // I don't understand why this case is
	       // is not handled automatically by the
               // QUOTED_STRING rules, but it is not.
             }

\'           BEGIN(QUOTED_STRING); break;

{date}       { yylval = make_Date(yytext);
               return yylval ? TOK_DATE : ERROR;
             }

\[           return L_LIST;

\]           return R_LIST;

\{           return L_SET;

\}           return R_SET;

\(           return L_PAREN;

\)           return R_PAREN;

.            ql_parse_error("Illegal token"); return ERROR;


<QUOTED_STRING,ESCAPED_STRING,ESCAPED_STRING1>\001   { 
            BEGIN(0);
            ql_parse_error("Query terminated prematurely by ^A in string");
            respond_with_control_A = 1; error_position.reset();
            return DOT;
          }


<QUOTED_STRING>[^'\n\\\001]+         {
            yylval = make_String(yytext);
            int r = tok_type((String*) yylval);
            return r;
          }

<QUOTED_STRING>[^'\n\\\001]*''        {
            BEGIN(ESCAPED_STRING); 
            yytext[yyleng - 1] = '\0';
            partial_string = (String*) make_String(yytext);
            if (!partial_string) return ERROR;
            break;
          }

<QUOTED_STRING>[^'\n\\\001]*\\ { 
          BEGIN(ESCAPED_STRING1);
          yytext[yyleng - 1] = '\0';
          partial_string = (String*) make_String(yytext);
          if (!partial_string) return ERROR;
          break;
        }

<ESCAPED_STRING>\\         { BEGIN(ESCAPED_STRING1);
                             yytext[yyleng - 1] = '\0';
                             partial_string = partial_string->put(yytext);
                             break;
                           }

<ESCAPED_STRING>''         { BEGIN(ESCAPED_STRING);
                             yytext[yyleng - 1] = '\0';
                             partial_string = partial_string->put(yytext);
                             break;
                           }

<ESCAPED_STRING1>[^\001]   { BEGIN(ESCAPED_STRING);
                          
                             if (yytext[0] <= 32 || (unsigned char) yytext[0] > 127)
                                break;  // Ignore escaped whitespace characters.
                             if (yytext[0] == '\n') error_position.nl();
                             switch (yytext[0]) {
                             case 'b': partial_string
                                         = partial_string->put("\b"); break;
                             case 't': partial_string
                                         = partial_string->put("\t"); break;
                             case 'n': partial_string
                                         = partial_string->put("\n"); break;
                             case '0':
                             case '1':
                             case '2':
                             case '3':
                                ql_parse_error
                                   ("Octal escapes not implemented yet");
                                break;
                             default: partial_string
                                        = partial_string->put(yytext); break;
                             // Many more cases to be handled.
                             }
                             break;
                           }


<ESCAPED_STRING>[^'\n\\\001]* partial_string=partial_string->put(yytext); break;


<ESCAPED_STRING>'       { BEGIN(0);
                          yylval = partial_string;
                          int r = tok_type(partial_string);
                          return r;
                        }


<QUOTED_STRING>'        BEGIN(0); break;

<QUOTED_STRING,ESCAPED_STRING>\n   { BEGIN(0);
                                     error_position.nl();
                                     ql_parse_error
                                       ("Unterminated quoted string, inserting single quote");
                                     break; }
