/* 
 * Copyright (c) 1995
 *       Whitehead Institute for Biomedical Research. All rights reserved.
 *
 * Released under terms specified in the Copyright file present in the top
 * directory of this distribution.
 */


#include "ql_dd.H"

#ifndef __TYPES_H
#include <sys/types.h>
#endif

#ifndef __STAT_H
#include <sys/stat.h>
#endif

#include <limits.h>

#ifndef DD_OPS_H
#include "dd_ops.H"
#endif

#ifndef QL_DD_ENTRY_H
#include "ql_dd_entry.H"
#endif

#ifndef QL_ERROR_H
#include "ql_error.H"
#endif

#ifndef MATERIAL_SET_H
#include "material_set.H"
#endif


QL_DD *QL_DD::_dd = 0;

QL_DD::QL_DD(/* const -- but Texas doesn't know it. */ char *database) :
  _hardwireds(0),_persistents(0),_database_file(database)
{
  ABORT(_dd,"Attempt to create duplicate data dictionary");
  _dd = this;

  // The following needed at least until
  // we re-set _persistents and other
  // database entry points after commit.
  int retain_persistent = 1;

#if defined(OBJECTSTORE)
  _transient_segment = os_segment::get_transient_segment();
#elif defined(TEXAS)
  _transient_segment = transient_store;
#else
  _transient_segment = NULL;
#endif

  if (!database) {
    _persistents = 0; _dd_root = 0; _tx = 0; _db = 0; 
    _next_tag_root = 0; _material_set_root = 0;
    // all segments are transient
    _step_segment = _material_segment = _tag_array_segment = _default_segment
      = _transient_segment;
  }
  else {
#if defined(OBJECTSTORE)
    struct stat buf;
    int r = stat(database, &buf);
    if (r) // failure
      {
	if (ENOENT == errno) _db = os_database::create(database);
	else ABORT(1,"Unable to find or create database");

	if (retain_persistent) objectstore::retain_persistent_addresses();

	_tx = os_transaction::begin(os_transaction::update);
	_dd_root = _db->create_root("data_dictionary");
        _next_tag_root = _db->create_root("next_tag");
        _material_set_root = _db->create_root("material_set");
	QL_DD_Entry **zero = new (_db) QL_DD_Entry*;
        *zero = 0;
	_dd_root->set_value(zero);
        short *next_tag = new (_db) short(1);
        _next_tag_root->set_value(next_tag);
        _material_set_root->set_value(new (_db) Value_Set_Table);
        _persistents = 0;
      }
    else {
      _db = os_database::open(database);
      if (retain_persistent) objectstore::retain_persistent_addresses();
      _tx = os_transaction::begin(os_transaction::update);
      _dd_root = _db->find_root("data_dictionary");
      _next_tag_root = _db->find_root("next_tag");
      _material_set_root = _db->find_root("material_set");
      QL_DD_Entry **entry = (QL_DD_Entry**) _dd_root->get_value();
      _persistents = *entry;
    }
    
    _default_segment = _db->get_default_segment();

    int segx;
    int n_segments = _db->get_n_segments();
    if (n_segments < 3) _step_segment = _db->create_segment();
    else {
      // WARNING!!!!!
      // we should really be storing the step segment index in
      // the database!!!
       _step_segment = _db->get_segment(4);
    }
    segx = _step_segment->get_number();
    ABORT(segx != 4, "Illegal segment number");

    n_segments = _db->get_n_segments();
    if (n_segments < 4) _tag_array_segment = _db->create_segment();
    else _tag_array_segment = _db->get_segment(6);
    segx = _tag_array_segment->get_number();
    ABORT(segx != 6, "Illegal segment number");

    n_segments = _db->get_n_segments();
    if (n_segments < 5) _material_segment = _db->create_segment();
    else _material_segment = _db->get_segment(8);
    segx = _material_segment->get_number();
    ABORT(segx != 8, "Illegal segment number");

    ABORT(_db->get_n_segments() != 5, "Fix this");

    // Debugging code:

    int transient_number = os_segment::get_transient_segment()->get_number();
    int default_number = _db->get_default_segment()->get_number();
    
    _db->set_check_illegal_pointers(0);  // For production.
    // _db->set_check_illegal_pointers(1);  // for debugging.
    //                  		    // very bad for commit performance
    step_segment()->null_illegal_pointers = 0;
    _db->default_segment->null_illegal_pointers = 0;
    os_transaction::commit();      
    _tx = os_transaction::begin(os_transaction::update);
#elif defined(TEXAS)
    p_store = _db = open_pstore(database);
    ABORT(!_db,"open_pstore returned 0");
    QL_DD_Entry **zero = pnew(_db,QL_DD_Entry*);
    ABORT(!zero,"pnew returned 0");
    *zero = NULL;
    // create and set the value of the _dd_root
    if (!is_root(_db,"data_dictionary")) {
       add_root(_db,QL_DD_Entry*,"data_dictionary",zero);
       _persistents = 0;
       _dd_root = (void*) get_root(_db,QL_DD_Entry*,"data_dictionary");
     } else {
      _dd_root = (void*) get_root(_db,QL_DD_Entry*,"data_dictionary");
      QL_DD_Entry **entry = (QL_DD_Entry**) _dd_root;
      _persistents = *entry;
    }
#undef ABORT
#define ABORT( condition, message ) \
if (condition) {\
 (cerr << "ABORT: " __FILE__ << ":" << __LINE__ << " " << message << endl).flush(); die_and_dump_core(); }

    if( !is_root(_db,"next_tag")) {
      int *next_tag = pnew(_db,int)(1);
      ABORT(!next_tag,"pnew returned 0 for next_tag");
      add_root(_db,int,"next_tag",next_tag);
    }
    _next_tag_root = (void*) get_root(_db,int,"next_tag");
    if( !is_root(_db,"material_set")) {
      Value_Set_Table *v_s_table = pnew(_db,Value_Set_Table);
      ABORT(!v_s_table,"pnew returned 0 for v_s_table");
      add_root(_db,Value_Set_Table,"material_set",v_s_table);
    }
    _material_set_root = (void*) get_root(_db,Value_Set_Table,"material_set");

#else
    ABORT(1,"You can not create a database under the main-memory version of labbase ");
#endif
  }

  // Possible problem w/ g++?
  ((DD_Ops*)0)->DD_Ops::add_hardwireds();
}


short QL_DD::get_and_increment_tag()
{
  static short next_tag_index = 1;  // Used when there is no database.
  short x;
  if (_dd->_db) {
#if defined(OBJECTSTORE)
    short *loc = (short*) _dd->_next_tag_root->get_value();
#elif defined(TEXAS)
    int *loc = (int*) _dd->_next_tag_root;
#else
    int *loc;      // this is a dummy declaration for use in MM version
                   // This branch is actually never called in MM version
    ABORT(1,
	  "(QL_DD::get_and_incremnt_tag)"
	  << "This branch should not be called if there is no database");
#endif
    ABORT(!loc, "NULL pointer gotten from _next_tag_root");
    x = (short)*loc;
    ABORT(x <= 0, "Too many tag indexes have been created");
    (*loc)++;
  } else {
    x = next_tag_index;
    next_tag_index++;
  }
  return x;
}



Value_Set_Table *QL_DD::persistent_m_sets()
{
  static Value_Set_Table *x = 0;
  if (!x) {
    if (_dd->_db) {
#if defined(OBJECTSTORE)
      x = (Value_Set_Table*) _dd->_material_set_root->get_value();
#elif defined(TEXAS)
      x = (Value_Set_Table*) _dd->_material_set_root;
#else
      ABORT(1,
	    "(QL_DD::persistent_m_sets)"
	    << "This branch should not be called if there is no database");
#endif
    } else x = new Value_Set_Table;
  }
  return x;
}


void QL_DD::add_persistent(QL_DD_Entry *e)
{ 
  ABORT(!e, "(QL_DD::add_persistent) Illegal argument")
  ABORT(find(e->identifier()),"Duplicate data dictionary entry");
  QL_DD_Entry **entry = 0;
  if (_dd->_dd_root) {
#if defined(OBJECTSTORE)
    entry = (QL_DD_Entry**) _dd->_dd_root->get_value();
    os_database *db1 = os_database::of(e);
    os_database *db2 = os_database::of(entry);
    ABORT(db1 != db2, "Wrong database");
#elif defined(TEXAS)
   entry = (QL_DD_Entry**) _dd->_dd_root;
#else
    ABORT(1,
	  "(QL_DD::add_persistent)"
	  << "This branch should not be called if there is no database");
#endif
  }
  e->set_next(_dd->_persistents);
  _dd->_persistents = e;
  if (entry) *entry = e;
}


QL_DD_Tag_Entry *QL_DD::find_by_tag_index(Tag tag)
{
  QL_DD_Entry *p;
  for (p = _dd->_hardwireds; p; p = p->next()) {
    if ((p->entry_kind() == QL_DD_Entry::TAG)
	&& ((QL_DD_Tag_Entry*)p)->tag_index() == tag)
      return (QL_DD_Tag_Entry*) p;
  }
  for (p = _dd->_persistents; p ;p = p->next()) {
    if ((p->entry_kind() == QL_DD_Entry::TAG)
	&& ((QL_DD_Tag_Entry*)p)->tag_index() == tag)
      return (QL_DD_Tag_Entry*) p;
  }
  return 0;
}


P_DATABASE *QL_DD::db()
{
#ifdef OBJECTSTORE
  if (_dd->_db) return _dd->_db;
  else return os_database::get_transient_database();
#else
    return _dd->_db;
#endif
}


P_SEGMENT *QL_DD::step_segment() { return _dd->_step_segment; }


P_SEGMENT *QL_DD::tag_array_segment() { return _dd->_tag_array_segment; }


P_SEGMENT *QL_DD::material_segment() { return _dd->_material_segment; }
  

P_SEGMENT *QL_DD::default_segment() { return _dd->_default_segment; }


P_SEGMENT *QL_DD::transient_segment() { return _dd->_transient_segment; }


void QL_DD::commit()
{
  if (_dd->_db) {
#if defined(OBJECTSTORE)
    os_transaction::commit();
    _dd->_tx = os_transaction::begin(os_transaction::update);
    begin_nested();
#elif defined(TEXAS)
   commit_transaction(_dd->_db);
#else
    ABORT(1,
	  "(QL_DD::commit)"
	  << "This branch should not be called if there is no database");
#endif
  }
}


#ifdef OBJECTSTORE
void QL_DD::begin_nested()
{
#ifdef NESTED_TXN
  if (_dd->_db) (void) os_transaction::begin(os_transaction::update);
#endif
}


void QL_DD::commit_nested()
{
#ifdef NESTED_TXN
  if (_dd->_db) os_transaction::commit();
#endif
}

void QL_DD::abort_nested()
{
#ifdef NESTED_TXN
  if (_dd->_db) os_transaction::abort();
#endif
}
#endif

void QL_DD::commit_last()
{
  if (_dd->_db)
#if defined(OBJECTSTORE)
 os_transaction::commit();
#elif defined(TEXAS)
  commit_transaction(_dd->_db);
#else
    ABORT(1,
	  "(QL_DD::commit_last)"
	  << "This branch should not be called if there is no database");
#endif
}

void QL_DD::abort()
{
  if (_dd->_db) {
#if defined(OBJECTSTORE)
     os_transaction::abort();
  _dd->_tx = os_transaction::begin(os_transaction::update);
#elif defined(TEXAS)
   abort_transaction(_dd->_db);
#else
    ABORT(1,
	  "(QL_DD::abort)"
	  << "This branch should not be called if there is no database");
#endif
   }
}


QL_DD_Entry *QL_DD::find(char *identifier)
{
  QL_DD_Entry *e = QL_DD_Entry::lookup(_dd->_hardwireds, identifier);
  if (e) return e;
  return QL_DD_Entry::lookup(_dd->_persistents, identifier);
}


QL_DD_Entry *QL_DD::find_or_error(char *identifier)
{
  QL_DD_Entry *e= find(identifier);
  if (!e) {
    ql_parse_error("Predicate '", identifier, "' is not known");
    return 0;
  } else return e;
}


QL_DD_Tag_Entry *QL_DD::tag_find(char *identifier)
{
  QL_DD_Entry *entry = find(identifier);
  if (!entry) return 0;
  if (QL_DD_Entry::TAG == entry->entry_kind())
    return (QL_DD_Tag_Entry*)entry;
  else
    return 0;
}


int QL_DD::db_size(off_t *bytes, long *blocks)
{
  ABORT(!bytes,"Need non-NULL argument (bytes)");
  ABORT(!blocks,"Need non-NULL argument (blocks)");
  struct stat buf;
  if (_dd->_db) {
    int r = stat(_dd->database_file(), &buf);
    if (r) return -1;
    *bytes = buf.st_size;
    *blocks = buf.st_blocks;
  } else {
    *bytes = 0;
    *blocks = 0;
  }
  return 0;
}
