/* 
 * Copyright (c) 1995
 *       Whitehead Institute for Biomedical Research. All rights reserved.
 *
 * Released under terms specified in the Copyright file present in the top
 * directory of this distribution.
 */


#include "ql_dd_entry.H"

#ifndef __STRING_H
#include <string.h>
#endif

#ifndef QL_ADORNMENT_MAP_H
#include "ql_adornment_map.H"
#endif

#ifndef QL_DD_H
#include "ql_dd.H"
#endif

#ifndef QL_ERROR_H
#include "ql_error.H"
#endif

#ifndef QL_TERM_H
#include "ql_term.H"
#endif

#ifndef QL_ARGS_H
#include "ql_args.H"
#endif

#ifndef ID_INDEX_ENTRY_H
#include "id_index_entry.H"
#endif

#if USE_TEXAS_INDEX && !defined(BTREE_H)
#include "btree.hh"
#endif

// Create a new data dictionary entry.
// The entry might have to be persistent unless it is
// the entry for a builtin.
QL_DD_Entry::QL_DD_Entry(char *identifier,char *who,Write_implementation ii,
			 Write_implementation di, Write_implementation ui) :
_next(0), _retired(0), _implementation_of_insert(ii),
_implementation_of_delete(di), _implementation_of_update(ui)
{
  _identifier = PNEW_ARRAY_ADDR(this,char,strlen(identifier) + 1);
  safe_strcpy(_identifier,identifier);
  _who = PNEW_ARRAY_ADDR(this,char,strlen(who) + 1);
  safe_strcpy(_who,who);
}


QL_DD_Builtin_Entry::QL_DD_Builtin_Entry(char *identifier,
					 Write_implementation ii,
					 Write_implementation di,
					 Write_implementation ui) :
  QL_DD_Entry(identifier,HARDWIRED_USER,ii,di,ui)
{
#ifdef OBJECTSTORE
  ABORT(os_database::of(this) != os_database::get_transient_database(),
        "Data dictionary entries for builtins should be transient");
#endif
    _adornment_map = new QL_Adornment_Map;
} 


QL_DD_Entry *QL_DD_Entry::lookup(QL_DD_Entry *E, char *identifier)
{
  while (E) {
    if (!strcmp(identifier, E->_identifier)) return E;
    E = E->_next;
  }
  return NULL;
}


QL_DD_Tag_Entry *QL_DD_Tag_Entry::new_dd_tag(char *identifer, Value::tv_type_id c_type,
					     Tag tag_index)
{
  QL_Type_Descriptor *td
    = c_type == Value::UNDEFINED ? 0 : new QL_Type_Descriptor(c_type);
  QL_DD_Tag_Entry *e
    = new QL_DD_Tag_Entry(identifer, tag_index, HARDWIRED_USER, td);
			  
  delete td;
  e->_dd_tag = 1;
  return e;  
}


static const int initial_id_index_size = 1000;

QL_DD_Tag_Entry::QL_DD_Tag_Entry(char *identifier, Tag tag_index, char *who,  
				  QL_Type_Descriptor *type, int create_only,
				  int id_tag, int segment, int put_on_history,
				  int associate_with_id)
:  QL_DD_Entry(identifier, who), _dd_tag(0), _create_only(create_only),
  _id_tag(id_tag), _segment_id(segment),  _put_on_history(put_on_history),
  _tag_index(tag_index), _associate_with_id(associate_with_id)
{
  ABORT(QL_DD::find_by_tag_index(_tag_index), "Duplicate tag index");
  // Copy the type descriptor into the database.
  _type = type ? PNEW_ADDR(this,QL_Type_Descriptor,(type)) : type;

  if (_id_tag) {
#if defined(TEXAS) && defined(USE_INDEX) 
    // this is needed because the comma in <K,V> messes up the macro
    typedef ID_INDEX_SET Index_Type;
    _id_index = PNEW_ADDR(this,Index_Type,);
#elif defined(MAIN_MEMORY)
    typedef ID_INDEX_SET Index_Type;
    _id_index = PNEW_ADDR(this,Index_Type,);
#endif
#if defined(OBJECTSTORE) && defined(USE_INDEX)
    _id_index->change_behavior(os_collection::pick_from_empty_returns_null);
    os_index_path &id_path
      = os_index_path::create("Id_Index_Entry*","_id",os_database::of(this));
    _id_index->add_index(id_path,
			 os_index_path::unordered
			 |os_index_path::no_duplicates
			 |os_index_path::signal_duplicates
			 |os_index_path::point_to_key);
#endif
  }
}


QL_DD_Entry::kind QL_DD_Material_Kind_Entry::entry_kind()
{ return MATERIAL_KIND; }


QL_DD_Material_Kind_Entry::QL_DD_Material_Kind_Entry(char *identifier,
						     char *who,
						     ID_INDEX_SET*main_index)
: QL_DD_Entry(identifier, who), _main_index(main_index), _supertype(0) {}


QL_DD_DD_Kind_Entry::QL_DD_DD_Kind_Entry(char *identifier,
					 Write_implementation ii,
					 Write_implementation di,
					 Write_implementation ui)
:  QL_DD_Entry(identifier, HARDWIRED_USER, ii, di, ui) {}


QL_DD_Step_Kind_Entry::QL_DD_Step_Kind_Entry(char *identifier, char *who)
:  QL_DD_Entry(identifier, who), _supertype(0) {}


char *QL_DD_Tag_Entry::primary_id_name(char *material_name)
{
  int len = strlen(material_name);
  char *output = new char [ len + 4];
  safe_strcpy(output, material_name);
  safe_strcpy(output + len, "_id");
  return output;
}


int QL_DD_Tag_Entry::check_duplicate(QL_Term *t, char *id)
{
  const Id_Index_Entry *e = lookup_material(t, id);
  if (e) { 
    ql_eval_error(t, "Duplicate id '", id, "' associated with tag '", identifier() ,"' in");
    return 0;
  } else return 1;
}


const Id_Index_Entry *QL_DD_Tag_Entry::lookup_material(QL_Term *, char *stack_id)
{

  static char this_id[2048];

  ABORT(strlen(stack_id) >= 2048, "Increase the size of this buffer");

  safe_strcpy(this_id, stack_id);

  const Id_Index_Entry *e;
#if  defined(OBJECTSTORE) && defined(USE_INDEX)

  static const os_coll_query* id_query = 0;
  static os_bound_query* bound_query = 0;
  if (!id_query) {
    id_query = &os_coll_query::create_pick("Id_Index_Entry*","!strcmp(_id,(char*)this_id)",
					  os_database::of(this),0);
    bound_query = new os_bound_query(*id_query,
				   ( os_keyword_arg("this_id",&this_id[0]) ));
  }
  e = id_index()->query_pick(*bound_query);
#elif USE_TEXAS_INDEX
  if (id_index() )  {     // if nonNULL index 
    String  *st = String::make(this_id);
    Str_Ptr sp1(st);
    e = id_index()->lookup(sp1);
  }
#else 
  // Do an iterative search.
  ID_INDEX_CURSOR *c = 
#ifdef OBJECTSTORE   // for ObjecStore without index
    id_index()->os_cursor(os_cursor::unsafe);
#else
    id_index()->cursor_of();
#endif
  for (e = c->first(); e; e = c->next())
    if (!strcmp(this_id, e->id())) {
      delete c;
      return e;
    }
  delete c;
#endif

  return e;
}


QL_Type_Descriptor::QL_Type_Descriptor(Value::tv_type_id x)
:_constructor(x), _arg_number(0), _args(0), _dd_constructor(0) {}


QL_Type_Descriptor::QL_Type_Descriptor(QL_Type_Descriptor *x)
:_constructor(x->constructor()), _arg_number(x->arg_number()),
 _dd_constructor(x->dd_constructor())
{
  _args = PNEW_ARRAY_ADDR(this,QL_Type_Descriptor*,arg_number());
  for (int i = 0; i<arg_number(); i++)
    _args[i] = PNEW_ADDR(this,QL_Type_Descriptor,(x->_args[i]));
}


QL_DD_Type_Constructor_Entry::QL_DD_Type_Constructor_Entry
(char *identifier, Value::tv_type_id type, int has_args)
:  QL_DD_Entry(identifier, HARDWIRED_USER),
   _has_args(has_args), _constructor(type), _named_type(0),
   _supertype(0) {}


QL_Type_Descriptor**
QL_Type_Descriptor::find_in_args(QL_Term *t, QL_Args *args, int *count)
{
  *count = args->arity();
  if (!*count) return 0;
  QL_Type_Descriptor **r = new QL_Type_Descriptor*[*count];
  for (int i = 0; args; args = args->next(), i++) {
    r[i] = new QL_Type_Descriptor(t, args);
    if (r[i]->constructor() == Value::UNDEFINED) {
      for (int j = 0; j <= i; j++) delete r[j];
      delete [] r;
      return 0;
    }
  }
  return r;
}


QL_Type_Descriptor::QL_Type_Descriptor(QL_Term *t, QL_Args *arg)
: _arg_number(0), _args(0), _constructor(Value::UNDEFINED),
  _dd_constructor(0)
{
  Value *tv = arg->value();
  if (!tv) {
    ql_eval_error(t, "Expect a bound value in");
    return;
  }
  QL_Term *type_term = tv2term(t,tv);
  if (!type_term
      || type_term->dd_entry()->entry_kind()
        != QL_DD_Entry::TYPE_CONSTRUCTOR) {
    // At this point one of 3 possible error conditions obtains:
    // 1) We have a string, but it is not the name of a type
    //    constructor (and was not changed to a term by tv2term):
    if (tv->tv_type() == Value::STRING) {
      char *chunk = tv->down_to_STRING()->get();
      ql_eval_error(t, "Unrecognized type '",  chunk, "' in");
      delete chunk;
    }
    // 2) We have a term, but the functor is not a constructor:
    else if (type_term) 
      ql_eval_error(t, "Unrecognized type '",
		    type_term->dd_entry()->identifier(),
		    "' in");
    // 3) We have neither a string nor a term, but some other value:
    else ql_eval_error
      (t, "Need a type identifier to be associated with tag 'type' in");
    return;
  }
  QL_DD_Type_Constructor_Entry *e
    = type_term->dd_entry()->down_to_TYPE_CONSTRUCTOR();

  if (e->constructor() == Value::UNDEFINED || e->constructor() == Value::TERM) {
    ql_eval_error(t, "Values of type '", e->identifier(),
		  "' cannot be stored in the database in");
    return;
  }

  int arity = type_term->args()->arity();
  if (!e->has_args() && arity) {
    ql_eval_error(t, "No arguments allowed for type '",
		  e->identifier(), "' in");
    return;
  }
  if (e->has_args()) {
    if (!arity) {
      ql_eval_error(t, "At least one argument required for type '",
		    e->identifier(), "' in");
      return; 
    }
    _args = find_in_args(t,type_term->args(),&_arg_number);
    if (!_args) return;
    if (e->constructor() != Value::TUPLE && arity > 1) {
      // There is an implicit TUPLE constructor, so we insert it here.
      QL_Type_Descriptor *tmp = new QL_Type_Descriptor(Value::TUPLE);
      tmp->_args = _args;
      tmp->_arg_number = _arg_number;
      _args = new QL_Type_Descriptor*;
      *_args = tmp;
      _arg_number = 1;
    }
  }
  _constructor = e->constructor();
}


QL_Type_Descriptor::~QL_Type_Descriptor()
{
  if (args()) for (int i = 0; i < arg_number(); i++) delete args()[i];
  delete args();
}


#define DEFINE_DOWNTO_FUNCTION(TYPE,RTYPE) \
RTYPE* QL_DD_Entry::CAT_ID2(down_to_,TYPE)() \
{ \
  ABORT(!this,"No entry found"); \
  ABORT(entry_kind() != TYPE,"Must be a RTYPE"); \
  return (RTYPE*) this; \
}

DEFINE_DOWNTO_FUNCTION(TAG,QL_DD_Tag_Entry)
DEFINE_DOWNTO_FUNCTION(BUILTIN,QL_DD_Builtin_Entry)
DEFINE_DOWNTO_FUNCTION(MATERIAL_KIND,QL_DD_Material_Kind_Entry)
DEFINE_DOWNTO_FUNCTION(STEP_KIND,QL_DD_Step_Kind_Entry)
DEFINE_DOWNTO_FUNCTION(DATA_DICTIONARY_KIND,QL_DD_DD_Kind_Entry)
DEFINE_DOWNTO_FUNCTION(TYPE_CONSTRUCTOR,QL_DD_Type_Constructor_Entry)
#undef DEFINE_DOWNTO_FUNCTION
