/* 
 * Copyright (c) 1995
 *       Whitehead Institute for Biomedical Research. All rights reserved.
 *
 * Released under terms specified in the Copyright file present in the top
 * directory of this distribution.
 */


// $Id: tag_query.C,v 1.3 1995/03/01 18:11:11 steve Exp $
#include "tag_query.H"

#ifndef	__sys_limits_h
#include <limits.h>
#endif

#ifndef QL_DD_ENTRY_H
#include "ql_dd_entry.H"
#endif

#ifndef QL_TERM_H
#include "ql_term.H"
#endif

#ifndef QL_ARGS_H
#include "ql_args.H"
#endif

#ifndef BASICS_H
#include "basics.H"
#endif

#ifndef VALUE_H
#include "value.H"
#endif

#ifndef ID_INDEX_ENTRY_H
#include "id_index_entry.H"
#endif

#ifndef MATERIAL_H
#include "material.H"
#endif

#ifndef STEP_H
#include "step.H"
#endif

#ifndef QL_ITERATOR_H
#include "ql_iterator.H"
#endif

#ifndef QL_BUILTIN_IMPL_H
#include "ql_builtin_impl.H"
#endif

class List_Value;

QL_Iterator *Tag_Query::create_iterator(QL_Term *t)
{
  QL_DD_Tag_Entry *e = t->dd_entry()->down_to_TAG();

  QL_Adornment a = t->adornment();
  
  if ((t->number_bound() + t->number_free()) < 2) {
    // This check could be relaxed to allow boolean
    // tags to operate as unary predicates.  There
    // is no downstream handling of this case at present,
    // however.
    ql_eval_error(t, "Require at least two arguments for");
    return 0;
  }

  if (e->id_tag() && 'b' == a[1])
    return new QL_Iterator(t, t->number_bound(), t->number_free(),
			   Tag_Query::id_lookup_first, 0);
  else {
    for (int i = 0; i < t->number_bound() + t->number_free() - 1; i++)
      if (a[i] != 'b') {
	ql_eval_error(t, "Each argument but the last must have a "
		      "'MATERIAL_POINTER' value in");
	return 0;
      }
    return new QL_Iterator(t, t->number_bound(), t->number_free(),
			   Tag_Query::first_N, 0);
  }
}


static QL_Iterator_State *
find_tag_in_step(QL_Term *t, QL_DD_Tag_Entry *entry,
		 Step *s,
                 int bound_number, Value *in[], int free_number, Value *out[])
{
  ABORT(free_number > 1, "This tag should have had arity 1");
  if (t); // Suppress warnings.
  Value *v = s->find(entry->tag_index());
  if (!v) return 0;
  if (free_number) { out[0] = v;  return (QL_Iterator_State*) 1; }
  else {
    if (*v == *in[bound_number-1]) return (QL_Iterator_State*) 1;
    else return 0;
  }
}


QL_FIRST_HEADER(Tag_Query::first_N)
{
  QL_DD_Tag_Entry *tag_entry = t->dd_entry()->down_to_TAG();

  int min_size = INT_MAX;

  if (m + n == 2 && in[0]->tv_type() == Value::STEP_POINTER)
    return find_tag_in_step(t, tag_entry, in[0]->down_to_STEP_POINTER(), m, in,
			    n, out);

  Material  *chosen_mpv = 0;

  List_Value* material_set = 0;

  if (m + n > 2) material_set = new List_Value(2);
  if (m + n < 2) {
    ql_eval_error(t, "Require at least two arguments for");
    return 0;
  }

  for (int i = 0; i < m + n -1; i++) {
    if (in[i]->tv_type() != Value::MATERIAL_POINTER) {
      ql_eval_error(t,"Need materials for all arguments but the last in");
      delete material_set;
      return 0;
    }
    Material *mpv = in[i]->down_to_MATERIAL_POINTER();

    if (m + n > 2) material_set->put(mpv);

    if (mpv->history_size() < min_size) {
      min_size = mpv->history_size();
      chosen_mpv = mpv;
    }
  }

  ABORT(!chosen_mpv, "Progamming error--no material chosen");

  if (m + n > 2) material_set->remove(chosen_mpv);

  Cursor_Step *c = chosen_mpv->step_cursor();
  if (!c) {
    ql_eval_error
      (t, "Possible system error; unable to obtain a cursor while evaluating");
    delete material_set;
    return 0;
  }

  Step *s; Value *v = 0;

  if (tag_entry->create_only()) {
    s = c->first();
    if (!s) {
      ql_eval_error(t, "No step history on material in");
      delete material_set;
      delete c;
      return 0;
    }
    v =s->find(tag_entry->tag_index());
  } else {
    for (s = c->last(); s; s = c->previous()) {
      v = s->find(tag_entry->tag_index());
      if (v && (m + n == 2 || s->contains_materials(material_set)))
	break; else v = 0;
    }
  }
  delete c;
  delete material_set;
  if (!v) return 0;
  if (n) { out[0] = v; return (QL_Iterator_State*) 1; }
  else {
    if (*v == *in[m-1]) return (QL_Iterator_State*) 1;
    else return 0;
  }
}


QL_FIRST_HEADER(Tag_Query::id_lookup_first)
{
  if (m + n != 2) {
    ql_eval_error(t, "Need exactly 2 arguments for id tag");
    return 0;
  }
  QL_DD_Tag_Entry *tag_entry = t->dd_entry()->down_to_TAG();
  Value *id_tv = n ? in[0] : in[1];
  if (id_tv->tv_type() != Value::STRING) {
    ql_eval_error(t, "The second argument must have type 'STRING' in");
    return 0;
  }
  char *id = id_tv->down_to_STRING()->get();
  const Id_Index_Entry *e = tag_entry->lookup_material(t,id);
  delete id;
  if (!e) return 0;
  if (n) {
    out[0] = e->material();
    return (QL_Iterator_State*) 1;
  } else {
    if (in[0]->tv_type() != Value::MATERIAL_POINTER) {
      ql_eval_error(t, "Require a 'MATERIAL_POINTER' as first argument in");
      return 0;
    }
    if (e->material() == in[0]->down_to_MATERIAL_POINTER())
      return (QL_Iterator_State*) 1; else return 0;
  }
}

