/* 
 * Copyright (c) 1995
 *       Whitehead Institute for Biomedical Research. All rights reserved.
 *
 * Released under terms specified in the Copyright file present in the top
 * directory of this distribution.
 */


//$Id: value.C,v 1.8 1995/03/01 13:37:25 steve Exp $

#include <errno.h>

#ifndef	__sys_limits_h
#include <limits.h>
#endif

#ifndef __TIME_H
#include <time.h>
#endif

#ifndef BASICS_H
#include "basics.H"
#endif

#ifndef VALUE_H
#include "value.H"
#endif

#ifndef __STRING_H
#include <string.h>
#endif

#ifndef SAFE_H
#include "safe.H"
#endif

#ifndef MATERIAL_H
#include "material.H"
#endif

#ifndef STEP_H
#include "step.H"
#endif

#ifndef QL_TERM_H
#include "ql_term.H"
#endif

#ifndef QL_ARGS_H
#include "ql_args.H"
#endif


// The following class gathers Value's
// into a list for deletion when GC::pop
// is called.

static void delete_value(Value*v);

class GC {
  class gnode {
  public:
    gnode(gnode * next, Value *v) : _next(next), _v(v) {}
    ~gnode() { delete_value(v()); }
    SLOT(Value*,v);
    SLOT(gnode*,next);
  };

  SLOT(gnode*,garbage_list);
  SLOT(GC*,next);
  static GC *lists;

public:
  static void add(Value*);
  static void push();
  static void pop();
  static void pop_last();
};


GC *GC::lists = 0;


void GC::pop_last()
{
  pop();
  ABORT(lists, "Lists should be 0");
} 
     

void GC::pop()
{
  ABORT(!lists, "Memory arena underflow");
  gnode *y = lists->garbage_list(), *y1;
  // Delete values first:
  while (y) {
    y1 = y;
    y = y->next();
    delete y1;
  }
  // Then do the the actual pop:
  GC *x = lists->next();
  delete lists;
  lists = x;
}


void GC::push()
{
  GC *x = new GC;
  x->_next = lists;
  x->_garbage_list = 0;
  lists = x;
}


void GC::add(Value *v)
{
  gnode *x = new gnode(lists->garbage_list(), v);
  lists->_garbage_list = x;
}




#define DEFINE_ZERO_OF_FN(TYPE,ARG,FN) \
  TYPE *TYPE::FN() { static TYPE x(ARG); return &x; }

#define DEFINE_ZERO_OF(TYPE,ARG) DEFINE_ZERO_OF_FN(TYPE,ARG,zero_of)
DEFINE_ZERO_OF(Int,"0")
DEFINE_ZERO_OF(Long_Int,"0")
DEFINE_ZERO_OF(Float,"0.0")
DEFINE_ZERO_OF(String,"")
DEFINE_ZERO_OF(Sequence,"")
DEFINE_ZERO_OF(Date,((time_t)0))
#undef DEFINE_ZERO_OF

DEFINE_ZERO_OF_FN(V_Collection,SET,set_zero_of)
DEFINE_ZERO_OF_FN(V_Collection,LIST,list_zero_of)
DEFINE_ZERO_OF_FN(V_Collection,SLIST,slist_zero_of)
#undef DEFINE_ZERO_OF_FN


Value *Value::zero_of(tv_type_id type)
{
  switch (type) {
  case INTEGER: return Int::zero_of();
  case FLOAT: return Float::zero_of();
  case STRING: return String::zero_of();
  case DNA_SEQUENCE: return Sequence::zero_of();
  case DATE: return Date::zero_of();
  case SET: return V_Collection::set_zero_of();
  case LIST: return V_Collection::list_zero_of();
  case SLIST: return V_Collection::slist_zero_of();
  case INTEGER32: return Long_Int::zero_of();
  default: return 0;
  }
}

void Value::pushmem() { GC::push(); }


void Value::popmem() { GC::pop(); }


void Value::popmem_last() { GC::pop_last(); }


Value::Value(tv_type_id the_type) : _type((char)the_type) {}


Value::~Value() {}


#define MAKE_GUTS GC::add(r); return r


String::String (const char* string) : Value(STRING), _data(0), _count(0)
// We must guarantee that, if the construction of ``this'' fails because
// len > USHRT_MAX, ~String will not be deleting invalid storage at
// _data.  Therefore we make _data 0 in the initializer.
{
  int len;	
  if (NULL == string) len = 1;
  else {
     len = strlen(string);
     if (len < 0) ABORT(1,errno_msg());
     len++;
  }
  if (len > USHRT_MAX) {
    set_error("String is too long");
    return;
  }
  _count = len;
  _data = PNEW_ARRAY_ADDR(this,char,count());
  safe_strcpy(_data, string);
}


String::String (const String* string)
: Value(STRING),
  _count(string != NULL ? string->count() : 1)
{
  _data = PNEW_ARRAY_ADDR(this,char,count());
  safe_strncpy(_data,string->_data,string->count());
  return ;
}


char String::get (const int i) const
{
  if (i < 0 || i > high ()) {
    ABORT(i<0,"String index out of bounds (negative) in String::get ()");
    ABORT(i>high(),"String index out of bounds (too big) in String::get ()");
  }
  return (_data[i]);
}


String *String::put(const char*src)
{
  int newlen = strlen(src), len = count() - 1;
  _count += newlen;
  char *temp = new char [count()];
  safe_strcpy(temp, _data);
  safe_strcpy(temp + len, src);
  delete [] _data;
  _data = temp;
  return this;
}


char*   String::get () const
{
  char* return_string = new char [count() + 1];
  safe_strncpy (return_string, _data, count());
  return_string[count()] = '\0';
  return (return_string);
}


int String::compare(const Value& v) const
{
  if (v.tv_type() != STRING) {
    set_error("Cannot compare a string to a non-string");
    return 1;
  }
  String *target = v.down_to_STRING();
  int cnt = min(count()-1,target->count()-1);
  for (int i = 0; i < cnt; i++)
    {
      if (get(i) < target->get(i))
	return -1;
      else if (get(i) > target->get(i))
	return +1;
    }
  // The strings are equal so far, but one or both may have run out
  //
  if (count() < target->count())
    return -1;
  else if (count() > target->count())
    return +1;
  else return 0;
}


String *String::substring(int start, int length) const
{
  clear_error();
  if ((start + length) > (int)(_count - 1) || (start < 0)) {
    set_error("Substring outside of string");
    return 0;
  }
  if (length < 0) {
    set_error("Length must be non-negative");
    return 0;
  }
  char *r = new char [ length + 1 ];
  for (int i = 0; i < length; i++) r[i] = get(i + start);
  r[i] = '\0';
  String *result = make(r);
  delete [] r;
  return result;
}


String *String::save(P_SEGMENT *s)
{ return PNEW(s,String,(this)); }


void String::print(ostream &os) const
{  os << _data; }


String *String::make(const char *s) { String *r = new String(s); MAKE_GUTS; }


Int::Int(long i) : Value(INTEGER), _int ((short)i)
{ 
  if (i < SHRT_MIN) {
    set_error("Underflow in creating 'INTEGER'");
    _int = -1;
  }
  if (i > SHRT_MAX) {
    set_error("Overflow in creating 'INTEGER'");
    _int = -1;
  }
}


Int::Int(const char* src) : Value(INTEGER), _int(-1)
{
  clear_error ();
  char *p;
  long x = strtol(src, &p, 10);
  if (p == src)  set_error("Cannot convert the string to an 'INTEGER'");
  else if (x < SHRT_MIN) set_error("Underflow in creating 'INTEGER'");
  else if (x > SHRT_MAX) set_error("Overflow in creating 'INTEGER'");
  else _int = (short) x;
}


Int* Int::make(const char*s) { Int *r = new Int(s); MAKE_GUTS; }


Int* Int::make(long i) { Int *r = new Int(i); MAKE_GUTS; }


void Int::print(ostream &os) const { os << _int; }


Int *Int::save(P_SEGMENT *s) { return PNEW(s,Int,(*this)); }


int Int::compare(const Value&a) const
{
  if (a.tv_type() != INTEGER && a.tv_type() != INTEGER32) {
    set_error("Must compare 'INTEGER' with 'INTEGER' or 'INTEGER32'");
    return 0;
  }
  long x = a.tv_type() == INTEGER
	   ? int(*a.down_to_INTEGER()) : int(*a.down_to_INTEGER32());
  return this->_int - x;
}


// Long_Int implementations:

Long_Int::Long_Int(long i) : Value(INTEGER32), _int (i) {}


Long_Int::Long_Int(const char* src) : Value(INTEGER32), _int(-1)
{
  clear_error ();
  char *p;
  long x = strtol(src, &p, 10);
  if (p == src)  set_error("Cannot convert the string to an 'INTEGER32'");
  _int = x;
}


Long_Int* Long_Int::make(long i) { Long_Int *r = new Long_Int(i); MAKE_GUTS; }


void Long_Int::print(ostream &os) const { os << _int; }


Long_Int *Long_Int::save(P_SEGMENT *s) { return PNEW(s,Long_Int,(*this)); }


int Long_Int::compare(const Value&a) const
{
  if (a.tv_type() != INTEGER && a.tv_type() != INTEGER32) {
    set_error("Must compare 'INTEGER32' with 'INTEGER' or 'INTEGER32'");
    return 0;
  }
  long x = a.tv_type() == INTEGER 
	   ? int(*a.down_to_INTEGER()) : int(*a.down_to_INTEGER32());
  return this->_int - x;
}


// Float implementations:

void Float::print(ostream &os) const { os << _float; }


Float *Float::save(P_SEGMENT *s) { return PNEW(s,Float,(*this)); }


int Float::compare(const Value&a) const
{
  if (a.tv_type() != FLOAT) {
    set_error("Must compare 'FLOAT' with 'FLOAT'");
    return 0;
  }
  double x = this->_float - a.down_to_FLOAT()->_float;
  if (x > 0.0) return 1;
  else if (x < 0.0) return -1;
  else return 0;
}


Float::Float(const char* src) : Value(FLOAT)
{
  clear_error ();
  char *p;
  errno = 0;
  _float = strtod(src,&p);
  if (p == src) {
    set_error("Cannot convert the string to a Float");
    _float = -1.0;
  } else if (errno) {
    set_error(errno_msg());	// This catches overflow.
    _float = -1.0;
  }
}


Float::Float(double d) : Value(FLOAT), _float(d) {}


Float *Float::make(const char* s) { Float *r = new Float(s); MAKE_GUTS; }


Float *Float::make(double d) { Float *r = new Float(d); MAKE_GUTS;  }


void Date::print(ostream &os) const
{
  struct tm* tmp = time_as_struct();
  if (tmp == NULL) return;
  char buffer[100];
  // I GIVE UP ON C++ IOS STATE VARIABLES!!
  sprintf(buffer,"%.4i:%.2i:%.2i:%.2i:%.2i:%.2i",tmp->tm_year+1900,tmp->tm_mon+1,
	  tmp->tm_mday,tmp->tm_hour,tmp->tm_min,tmp->tm_sec);
  os << buffer;
}


Date *Date::make(time_t clock) { Date *r = new Date(clock); MAKE_GUTS; }


Date *Date::make(const char* s) { Date *r = new Date(s); MAKE_GUTS; }


Date *Date::save(P_SEGMENT *s) { return PNEW(s,Date,(*this)); }


int Date::compare(const Value& a) const
{
  if (a.tv_type() != DATE) {
    set_error("Must compare 'DATE' with 'DATE'");
    return 1;
  }
  return this->_clock - a.down_to_DATE()->_clock;
}


extern "C" { char *strptime(const char *, const char *, struct tm*); }

Date::Date(const char* s) : Value(DATE), _clock(0)
{
  static struct tm zero_tm = { 0 };
  struct tm tm = zero_tm;

  // The date format is YYYY:MM:DD:HH:MM:SS
  const char *last = strptime(s, "%Y:%m:%e:%T", &tm);
  // strptime is non-POSIX.  Replace the call above with
  // a call to sscanf and appropriate error checking and
  // adjustment of the elements of tm.  Rethink the policy
  // for whether the date must be complete; if it must be,
  // enforce it.
  if (!last) {
    set_error("Illegal date (some fields may be missing or may have illegal values)");
    _clock = 0;
  } else if (last - s != strlen(s)) {
    set_error("Too much input");
    _clock = 0;
  } else _clock = timelocal(&tm);
}


#define DEFINE_DOWNTO_FUNCTION(TYPE,RTYPE) \
RTYPE* CAT_ID2(Value::down_to_,TYPE()) const\
{ \
    ABORT(!this,"No entry found"); \
      ABORT(tv_type() != TYPE,"Must be a TYPE"); \
	return (RTYPE*) this; \
	}

DEFINE_DOWNTO_FUNCTION(STRING,String)
     DEFINE_DOWNTO_FUNCTION(INTEGER,Int)
     DEFINE_DOWNTO_FUNCTION(FLOAT,Float)
     DEFINE_DOWNTO_FUNCTION(DATE,Date)
     DEFINE_DOWNTO_FUNCTION(DNA_SEQUENCE,Sequence)
     DEFINE_DOWNTO_FUNCTION(LIST,V_Collection)
     DEFINE_DOWNTO_FUNCTION(SET,V_Collection)
     DEFINE_DOWNTO_FUNCTION(MATERIAL_POINTER,Material)
     DEFINE_DOWNTO_FUNCTION(STEP_POINTER,Step)
     DEFINE_DOWNTO_FUNCTION(TERM,QL_Term)
     DEFINE_DOWNTO_FUNCTION(SLIST,V_Collection);
     DEFINE_DOWNTO_FUNCTION(INTEGER32,Long_Int);
#undef DEFINE_DOWNTO_FUNCTION
     
     
#define TRY_TYPE( type ) if (!strcmp(a,MACRO_STRING(type))) { return type; } 
Value::tv_type_id Value::char2type (const char* a)
{
  TRY_TYPE(STRING)
  TRY_TYPE(INTEGER)
  TRY_TYPE(FLOAT)
  TRY_TYPE(DATE)
  TRY_TYPE(BOOLEAN)
  TRY_TYPE(LIST)
  TRY_TYPE(SET)
  TRY_TYPE(MATERIAL_POINTER)
  TRY_TYPE(DNA_SEQUENCE)
  TRY_TYPE(STEP_POINTER)
  TRY_TYPE(TERM)
  TRY_TYPE(SLIST)
  TRY_TYPE(TUPLE)
  TRY_TYPE(INTEGER32)
  return UNDEFINED;
}
#undef TRY_TYPE


#define TRY_NAME( type ) if (thetype == type) { return MACRO_STRING(type); } 
const char* Value::type2char(tv_type_id thetype)
{
  TRY_NAME(STRING)
  TRY_NAME(INTEGER)
  TRY_NAME(FLOAT)
  TRY_NAME(DATE)
  TRY_NAME(BOOLEAN)
  TRY_NAME(LIST)
  TRY_NAME(SET)
  TRY_NAME(MATERIAL_POINTER)
  TRY_NAME(DNA_SEQUENCE)
  TRY_NAME(STEP_POINTER)
  TRY_NAME(TERM)
  TRY_NAME(SLIST)
  TRY_NAME(TUPLE)
  TRY_NAME(INTEGER32)
  return "UNDEFINED";
}
#undef TRY_NAME


// There are two implementations of V_Cursor, each
// implemented as a derived class.

class Sparse_List_Element : public Value  // Element of a sparse list.
// Derived from Value so that we can simply store
// them in the _collection slot of V_Collection.
// This choice is a hack designed to avoid disrupting
// of the Value clients at this point.  Also,
// reoganizing the V_Collection hierarchy could now
// could be wasted effort w.r.t. the Theta version of Labbase.
{
public:
  static Sparse_List_Element *make(short i, Value *v);
  // Deletion of the Value * _should_not_ occur in the
  // destructor of Sparse_List_Element; we use the default destructor.
  Value *save(P_SEGMENT*);
  void unsave();
  SLOT(short,index);
  SLOT(Value*,v);
protected:
  Sparse_List_Element(short i, Value *v) : Value(SLIST_ELEM),_index(i),_v(v) {}
};


Sparse_List_Element * Sparse_List_Element::make(short i, Value *v)
{
  Sparse_List_Element *r = new Sparse_List_Element(i,v); MAKE_GUTS;
}


Value *Sparse_List_Element::save(P_SEGMENT *s)
{
  Sparse_List_Element *r = PNEW(s,Sparse_List_Element,(index(),v()->save(s)));
  return r;
}


void Sparse_List_Element::unsave()
{
  _v->unsave();
  delete this;
}


V_Cursor::~V_Cursor() {}

class V_Cursor_Sparse : public V_Cursor // For sparse lists.
{
public:
  Value *first();
  Value *next();
  V_Cursor_Sparse(const V_Collection* collection) : c(collection),
				   _current_index(0), _data_index(0) {}
  const Sparse_List_Element * next_element() const { return _next_element; }
protected:
  SLOT(short,current_index);
  SLOT(short,data_index);
  SLOT(short,number_elements);
  const Sparse_List_Element* _next_element;
  SLOT(Value*,zero_Value);
  Value *get_next();
  const V_Collection* c;
};


// The representation of sparse lists is as follows:
// The first element in the list is the '0' Value:
// the Value for ``missing'' elements.
// The index of the '0' value is the _number_ of 
// values in the array.


Value *V_Cursor_Sparse::first()
{
  ABORT(c->count()==0,"No elements allocated for SLIST in V_Collection");
  _current_index = 0;
  _data_index = 0;
  Value *v = c->_data[_current_index];

  ABORT(!v,"Require at least one element to store the 0-Value");
  ABORT(v->tv_type() != Value::SLIST_ELEM, "Need a Sparse_List_Element");
  Sparse_List_Element *e = (Sparse_List_Element*) v; 
  // Deliberately do not use a down_to... function here to avoid
  // crudding up Value's public interface.

  _number_elements = e->index();
  _zero_Value = e->v();

  // set _next_element to the next nonzero elements if it exists
  _data_index++;
  if ( data_index() < c->count() ) {
    Value *v =  c->_data[data_index()];
    ABORT(v && v->tv_type() != Value::SLIST_ELEM,
	       "Need a Sparse_List_Element (2)");
    _next_element = (Sparse_List_Element*) v;
    ABORT(next_element()->index() >= number_elements(),
       "Index larger than total # elements");

  } else
    _next_element = NULL;
  return next();
}


Value *V_Cursor_Sparse::next() 
// Before calling, _next_element should be set
// to the next element in the spare representation of the list.
// This invariant is maintained by next().
{
  ABORT(!zero_Value(), "V_Cursor_Sparse::first() not called");

  // return null when end of the original list is reached
  if (current_index() >= number_elements() ) return NULL;

  Value *v = zero_Value();
  if (next_element()) {
    ABORT(next_element()->index() < current_index(), "current_index too large");
    // return a nonzero element if its index equals _current_index
    if (next_element()->index() == current_index()) {
      v = next_element()->v();
      // set next_element to the next SLIST element if it exists
      _data_index++;
      if (data_index() < c->count()) {
        Value *temp = c->_data[data_index()]; 
        ABORT(temp && temp->tv_type() != Value::SLIST_ELEM,
	    "Need a Sparse_List_Element (3)");
        _next_element = (Sparse_List_Element*) temp;
        ABORT(next_element()->index() >= number_elements(),
	   "Index larger than total # elements");
      } else // it's null if no more nonzero elements are left
        _next_element = NULL;	
    }
  }
  _current_index++;
  return v;
}


class V_Cursor_STD : public V_Cursor 
{
public:
  Value *first();
  Value *next();
  V_Cursor_STD(const V_Collection *collection):c(collection),_cursor_index(0) {}
protected:
  short _cursor_index;
  const V_Collection *c;
};

Value *V_Cursor_STD::first()
{
  _cursor_index = 0;
  Value *r = c->count() > 0 ? c->_data[_cursor_index]: NULL;
  _cursor_index++;
  return r;
}

Value *V_Cursor_STD::next()
{
  Value *r = (_cursor_index < c->count())?c->_data[_cursor_index] : NULL;
  _cursor_index++;
  return r;
}


V_Collection::V_Collection(tv_type_id type, int size): Value(type), _data(0), 
_count(0)
// We must guarantee that, if the construction of ``this'' fails because
// size > USHRT_MAX, ~V_Collection will not be deleting invalid storage at
// _data.  Therefore we make _data 0 in the initializer.
{
  if (size > USHRT_MAX) {
    set_error("Collection would be too big");
    return;
  }
  ABORT(size < 0," Negative size for the array in V_Collection ");
  _count = size;
  _data = PNEW_ARRAY_ADDR(this,Value*,count());
}


V_Collection *V_Collection::save(P_SEGMENT *seg)
{
  V_Collection *r = PNEW(seg,V_Collection,(tv_type(), count()));

  for (int i = 0; i < count(); i++)
    r->put(i, _data[i]->save(seg));

  return r;
}


void V_Collection::unsave()
{
  for (int i = 0; i < count(); i++) _data[i]->unsave();
  delete this;
}


V_Collection *V_Collection::make(tv_type_id type, const QL_Args *p)
{
  ABORT(type == SLIST && p, "Cannot make an SLIST from QL_Args");
  V_Collection *r = new V_Collection(type, p->arity());
  if (errmsg()) { delete r; return NULL; }
  GC::add(r);
  int index=0;
  for ( ; p; p = p->next()) {
    Value *v = p->value();
    ABORT(!v, "The parser should not allow this");
    r->put(index++,v);
  }
  return r;
}


V_Collection *V_Collection::make(tv_type_id type, const V_Collection* list,
				 Value *zero_Value)
{
  ABORT(type != SLIST, "This function creates an SLIST only");
  ABORT(list->tv_type() != LIST, "This function requires a LIST as second argument");
  V_Cursor *c = list->cursor_of();
  Value *v = c->first();
  ABORT(v && v->tv_type() != zero_Value->tv_type(),
	"Wrong tv_type() for zero_Value");

  // count the number of nonzero elements in the list
  int NonZeros = 0;
  for (; v; v = c->next())
    if (*v != *zero_Value)
      NonZeros++;

  V_Collection *r = new V_Collection(type, NonZeros+1);
  if (errmsg()) { delete r; delete c; return NULL; }
  GC::add(r);

  // Store the cardinality of the original list in the zero-th position
  // of _data array.
  r->put(0, Sparse_List_Element::make( list->cardinality(),zero_Value));

  // Enter the nonzero elements of the list 
  v = c->first();
  for (int i = 1, j = 0; v; v = c->next(), j++)
    if (*v != *zero_Value) {
      r->put(i, Sparse_List_Element::make(j,v));
      i++;
    }

  delete c;
  return r;
}


V_Collection *V_Collection::make(tv_type_id type, const List_Value* in)
{
    ABORT(type != LIST && type != SET, "Valid only for LIST or SET");
    V_Collection *r = new V_Collection(type, in->count());
    if (errmsg()) { delete r; return NULL; }
    GC::add(r);
    Cursor_Value* c = in->cursor_of();
    int i = 0;
    for (Value *v = c->first(); v; v = c->next(), i++) r->put(i,v);	    
    delete c;
    return r;
}


V_Collection *
V_Collection::make(Value::tv_type_id type, unsigned nelems, Value*in[])
{
   ABORT(type != LIST && type != SET, "Valid only for LIST or SET");
   V_Collection *r = new V_Collection(type, nelems);
   if (errmsg()) { delete r; return NULL; }
   GC::add(r);
   for (int i = 0; i < nelems; i++) r->put(i,in[i]);
   return r;
}


V_Cursor *V_Collection::cursor_of() const
{
  // The underlying objectstore cursor is _UNSAFE_.
  // One must not insert to or delete from the collection while
  // traversing it.

  if (tv_type() == SLIST)
    return new V_Cursor_Sparse(this);
  else
    return new V_Cursor_STD(this);
}


void V_Collection::put(int index, Value* v)
{
  // Check for out of bounds index values 
  ABORT(index < 0, "Index < 0");
  ABORT(index >= count(), "Index >= count()");
  _data[index] = v;
  
}
void V_Collection::print(ostream &os) const
{
  char *start, *end;
  
  if (tv_type() == LIST || tv_type() == SLIST) { start = "["; end = "]"; }
  else if (tv_type() == SET) { start = "{"; end = "}"; }
  else ABORT(1, "Illegal type tag");

  os << start;
  V_Cursor*c = cursor_of();
  char *comma = "";
  for (Value *v =  c->first(); v; v = c->next()) {
    os << comma;
    os << v;
    comma = ",";
  }
  delete c;
  os << end;
}


int V_Collection::compare(const Value &v) const
{
  if (tv_type() == LIST || tv_type() == SLIST) {
    if (v.tv_type() == LIST || v.tv_type() == SLIST) {
      V_Collection* list;
      if (v.tv_type() == LIST) list = v.down_to_LIST();
      else if (v.tv_type() == SLIST) list = v.down_to_SLIST();
      V_Cursor *c1 = cursor_of();
      V_Cursor *c2 = list->cursor_of();
      for (Value *v1 = c1->first(), *v2 = c2->first();
	   v1 && v2;
	   v1 = c1->next(), v2 = c2->next()) {
	int diff = v1->compare(*v2);
	if (errmsg()) { delete c1; delete c2; return 1; }
	if (diff) { delete c1; delete c2; return diff; }
      }
      delete c1; delete c2;
      if (!v1 && !v2) return 0;
      else if (!v1 && v2) return -1;
      else if (v1 && !v2) return 1;
      else ABORT(1,"Programming error in V_Collection::compare");
    }
    else
      set_error("Must compare 'LIST' or 'SLIST' to 'LIST' or 'SLIST'");
  }
  else if (tv_type() == SET) {
    if (v.tv_type() == SET)
      set_error("'SET' comparisions not implemented");
    else set_error("Must compare 'SET' to 'SET'");
  }
  else ABORT(1, "Illegal type");
  return 1;
}


Value *V_Collection::get(int index) const
{

  ABORT((index < 0) || (index >= cardinality()),
		  "Index out of range, should have been checked by the caller");
  if (tv_type() == LIST) return _data[index];
  else if (tv_type() == SLIST) {
    V_Cursor *c = cursor_of();
    Value *v = c->first();
    for (int i = 0; (i != index) && v; v = c->next(), i++); 

    ABORT(!v,"Index out of range, should have been checked by the caller");
    delete c;
    return v;
  }
  else {
    ABORT(1,"Get is supported only on Lists");
    return 0;
  }
}


int V_Collection::cardinality() const
{
  if (tv_type() == LIST || tv_type() == SET)
    return count();
  else if (tv_type() == SLIST) {
    /* count is always set to 1 + #Of Elements */
    ABORT(count() < 1,"Underlying list must have at least one element");
    ABORT(_data[0]->tv_type() != SLIST_ELEM,
	  "Elements must have type Sparse_List_Element");
    return ((Sparse_List_Element*)_data[0])->index();
  }
}


// int V_Collection::empty() const
// {
//   return cardinality() == 0;
// }


Value::tv_type_id Value::tv_type () const
{ 
  return (tv_type_id) _type;
}



# define V_PRINT(TYPE) \
 case CAT_ID2(Value::,TYPE): v.CAT_ID2(down_to_,TYPE())->print(os); break
 ostream &operator << (ostream &os, const Value &v) 
{
  switch (v.tv_type()){
  case Value::UNDEFINED: ABORT(1, "Cannot print undefined Value");
			 V_PRINT(DATE);
			 V_PRINT(INTEGER);
			 V_PRINT(STRING);
			 V_PRINT(FLOAT);
			 V_PRINT(LIST);
			 V_PRINT(SET);
			 V_PRINT(MATERIAL_POINTER);
			 V_PRINT(STEP_POINTER);
			 V_PRINT(TERM);
			 V_PRINT(DNA_SEQUENCE);
			 V_PRINT(SLIST);
//			 V_PRINT(TUPLE);
                         V_PRINT(INTEGER32);
		       default: ABORT(1, "Uknown branch in Value::print");
		       }
  return os;
}
#undef V_PRINT			

ostream &operator << (ostream &os, const Value *v)
{
  return (os << *v);
}


# define V_SAVE(TYPE) \
 case CAT_ID2(Value::,TYPE): return this->CAT_ID2(down_to_,TYPE())->save(seg)
Value *Value::save (P_SEGMENT *seg)
{
  switch (this->tv_type()){
  case UNDEFINED: ABORT(1, "Cannot save UNDEFINED Value");
  V_SAVE(DATE);
  V_SAVE(INTEGER);
  V_SAVE(STRING);
  V_SAVE(FLOAT);
  V_SAVE(LIST);
  V_SAVE(SET);
  case MATERIAL_POINTER: return this; // Already in db.
  case STEP_POINTER: ABORT(1,"Cannot save steps");
  case TERM: ABORT(1, "Cannot (yet) save terms");
  V_SAVE(DNA_SEQUENCE);
  V_SAVE(SLIST);
  case SLIST_ELEM: return ((Sparse_List_Element*)this)->save(seg);
  V_SAVE(INTEGER32);
  default: ABORT(1, "Uknown branch in Value::save");
  }
}
#undef V_SAVE

#define V_UNSAVE(TYPE) case TYPE: delete this->CAT_ID2(down_to_,TYPE()); return
void Value::unsave() {
  switch (this->tv_type()) {
  case UNDEFINED:  ABORT(1, "Cannot unsave UNDEFINED Value");
  V_UNSAVE(DATE);
  V_UNSAVE(INTEGER);
  V_UNSAVE(STRING);
  V_UNSAVE(FLOAT);
  case LIST: this->down_to_LIST()->unsave(); return;
  case SET: this->down_to_SET()->unsave(); return;
  case MATERIAL_POINTER:
  case STEP_POINTER:
    ABORT(1, "Illegal argument to Value::unsave");
  V_UNSAVE(DNA_SEQUENCE);
  case SLIST: this->down_to_SLIST()->unsave(); return;
  case SLIST_ELEM: ((Sparse_List_Element*)this)->unsave(); return;
  V_UNSAVE(INTEGER32);
  default: ABORT(1, "Unknown branch in Value::unsave()");
  }
}
#undef V_UNSAVE


# define V_COMPARE(TYPE) \
 case CAT_ID2(Value::,TYPE): return this->CAT_ID2(down_to_,TYPE())->compare(v)
int Value::compare(const Value&v) const
{
  clear_error();
  switch (this->tv_type()){
  case Value::UNDEFINED: ABORT(1, "Cannot compare undefined Value");
			 V_COMPARE(DATE);
			 V_COMPARE(INTEGER);
			 V_COMPARE(STRING);
			 V_COMPARE(FLOAT);
			 V_COMPARE(LIST);
			 V_COMPARE(SET);
			 V_COMPARE(MATERIAL_POINTER);
			 V_COMPARE(STEP_POINTER);
			 V_COMPARE(TERM);
			 V_COMPARE(DNA_SEQUENCE);
                         V_COMPARE(SLIST);
                         // V_COMPARE(TUPLE);
                         V_COMPARE(INTEGER32);
		       default: ABORT(1, "Uknown branch in Value::compare");
		       }
}
#undef V_COMPARE


int operator == (Value const &v1, Value const &v2)
{ return !v1.compare(v2); }

int operator != (Value const &v1, Value const &v2)
{ return !!v1.compare(v2); }

int operator <= (Value const &v1, Value const &v2)
{ return v1.compare(v2) <= 0; }

int operator >= (Value const &v1, Value const &v2)
{ return v1.compare(v2) >= 0; }

int operator < (Value const &v1, Value const &v2)
{ return v1.compare(v2) < 0; }

int operator > (Value const &v1, Value const &v2)
{ return v1.compare(v2) > 0; }


Sequence* Sequence::reverse_complement(int start, int len) const
{
  clear_error();
  if (start > high() || start + 1 < len) {
    set_error("Sub sequence not contained in sequence in Sequence::reverse_complement");
    return 0;
  }
  char *r = new char [len + 1], *j = r;
  for ( ; len>0; start--, len--, j++) *j = reverse_complement(get(start));
  *j = '\0';
  Sequence *result = make(r);
  delete [] r;
  return result;
}


Sequence* Sequence::reverse_complement() const
{
  return reverse_complement(high(),count() - 1);
}


char Sequence::reverse_complement(char a)
{
  switch (a) {
  case 'g': case 'G':
    return 'C';
    
  case 'c': case 'C':
    return 'G';
    
  case 'a': case 'A':
    return 'T';
    
  case 't': case'T':
    return 'A';
    
  default:
    return 'N';
  }
}


int Sequence::valid() const
{
  for (int i = 0; i<=high(); i++) {
    char a = get(i);
    switch (a) {
    case 'a': case 'g':
    case 'c': case 't':
    case 'n': case 'A':
    case 'G': case 'C':
    case 'T': case 'N':
      continue;
    default: return 0;
    }
  }
  return 1;
}


Sequence::Sequence (const String *in) : String(in) { _type = (char) DNA_SEQUENCE; }


Sequence *Sequence::make(const String *in)
{ Sequence *r = new Sequence(in); MAKE_GUTS; }
  

Sequence::Sequence (const char *in) : String(in) { _type = (char) DNA_SEQUENCE; }


Sequence *Sequence::make(const char *in)
{ Sequence *r = new Sequence(in); MAKE_GUTS; }


static void delete_value(Value*v)
{
  if (!v) return;		// To be consistent with C++'s delete operator.
  switch (v->tv_type()) {
  case Value::STRING:
    delete v->down_to_STRING(); break;
  case Value::DNA_SEQUENCE:
    delete v->down_to_DNA_SEQUENCE(); break;
  case Value::LIST: delete v->down_to_LIST(); break;
  case Value::SET: delete v->down_to_SET(); break;
  case Value::FLOAT:
  case Value::INTEGER:
  case Value::DATE:
  case Value::INTEGER32: delete v; break;
  case Value::TERM:		// Right now terms are deleted in a special way.
  case Value::STEP_POINTER: case Value::MATERIAL_POINTER:
      ABORT(1,"Should not delete these");
  case Value::SLIST: delete v->down_to_SLIST(); break;
  case Value::SLIST_ELEM: delete (Sparse_List_Element*) v; break;
  // case Value::TUPLE: delete v->down_to_TUPLE(); break;
  case Value::UNDEFINED: ABORT(1, "Cannot destroy an undefined value"); break;
  default: ABORT(1,"Illegal tv_type_id");
  }
}
