/* 
 * Copyright (c) 1995
 *       Whitehead Institute for Biomedical Research. All rights reserved.
 *
 * Released under terms specified in the Copyright file present in the top
 * directory of this distribution.
 */


#ifndef VALUE_H   // first line of file
#define VALUE_H

#ifndef __TIME_H
#include <time.h>
#endif

#ifdef OBJECTSTORE
#ifndef _OBJECTSTORE_H_
#include <objectstore/objectstore.H>
#endif

#ifndef _OS_COLLECTION_H_
#include <ostore/coll.hh>
#endif
#endif

#ifndef SAFE_H
#include "safe.H"
#endif

#ifndef LIST_H
#include "list.H"
#endif

#ifndef __CTYPE_H
#include "ctype.h"
#endif


// This file defines class Value, its derived classes, and an auxiliary class,
// V_Cursor, for iterating over elements of the collection class derived from
// Value.  A key consideration in the definition of Value and its derived
// classes is the desire to save space.  For this reason Value does not use
// virtual functions, even though they would be natural.  Instead, the function
// in Value in many cases devolves to a dispacth to the correct function, based
// on the explicit run-time type descriptor stored in the value.  The desire to
// save space also sometimes governs the order in which data members are defined
// within the classes derived from Value.
//
// Not all classes derived from Value are defined in this file: the notable
// exceptions are classes Step, QL_Term, and Material.


// Declare subclasses of Value:
class String;
class Int;
class Float;
class Date;
class V_Collection;
class Material;
class Step;
class QL_Term;
class Sequence;
class Long_Int;

// Declare other classes:
class List_Value;

class Value 
{
 public:

  // The following enum provides values for run-time type descriptors of Value
  // instances.
  enum tv_type_id {
    UNDEFINED=0,
    DATE=1,
    INTEGER=2,
    BOOLEAN=3,
    STRING=4,
    FLOAT=5,
    LIST=6,
    SET=7,
    MATERIAL_POINTER=8,
    STEP_POINTER=9,
    TERM=10,
    DNA_SEQUENCE=11,
    SLIST=12,
    SLIST_ELEM=13, // NOT FOR PUBLIC CONSUMPTION.
    INTEGER32=14,
    TUPLE=15,
    // Extra types, to possible avoid need future need for database reloads.
    EXTRA_TYPE1=16,
    EXTRA_TYPE2=17,
    EXTRA_TYPE3=18,
    };
  
  
  Value(tv_type_id);
  ~Value();

  tv_type_id tv_type() const;

  // "save" copies a value and all its sub-values into persistent storage.
  Value *save(P_SEGMENT *);
  // 'unsave" deletes the storage allocated by "save".  It is an error to call
  // "unsave" on a MATERIAL_POINTER or STEP_POINTER (Material or Step).
  void unsave();

  int compare(const Value&) const;

  String *down_to_STRING() const;
  Int *down_to_INTEGER() const;
  Float *down_to_FLOAT() const;
  Date *down_to_DATE() const;
  V_Collection *down_to_LIST() const;
  V_Collection *down_to_SET() const;
  Material *down_to_MATERIAL_POINTER() const;
  Step *down_to_STEP_POINTER() const;
  QL_Term *down_to_TERM() const;
  Sequence *down_to_DNA_SEQUENCE() const;
  V_Collection *down_to_SLIST() const;
  Long_Int *down_to_INTEGER32() const;
  V_Collection *down_to_TUPLE() const;

  static tv_type_id char2type (const char *a);
  static const char* type2char (tv_type_id thetype);

  static Value* zero_of(tv_type_id type); // Return the 0 Value for type.
					  // The return value should really
					  // be 'const'.

  static void pushmem();     // Start a new ``arena'' for Value's
  static void popmem();      // Reclaim memory from the most
			     // recent ``arena'' and make the
			     // previous area the most recent.
  static void popmem_last(); // Like popmem(), but checks that
			     // all memory has been poped.

protected:
  char _type;
};


ostream& operator<< (ostream& os, const Value* tv);
ostream& operator<< (ostream& os, const Value& tv);

int operator== (const Value& a,const Value& b);
int operator<= (const Value& a,const Value& b);
int operator<  (const Value& a,const Value& b);
int operator!= (const Value& a,const Value& b);
int operator>  (const Value& a,const Value& b);
int operator>= (const Value& a,const Value& b);


class Int : public Value
{
public:
  static Int *make(const char*src);
  static Int *make(long i);
  
  operator int () const { return _int; }
  
  void print(ostream&) const;
  Int *save(P_SEGMENT *dbs);
  int compare(const Value&) const;
  static Int *zero_of();

protected:
  Int (long i);
  Int(const char *src);
  short _int;
};


// Long_Int is intended to be identical to
// Int, except that it will be able to hold
// C long int values, not just short int values.
class Long_Int : public Value
{
public:
  static Long_Int *make(const char*src);
  static Long_Int *make(long i);
  
  operator long int () const { return _int; }
  
  void print(ostream&) const;
  Long_Int *save(P_SEGMENT *dbs);
  int compare(const Value&) const;
  static Long_Int *zero_of();

protected:
  Long_Int (long i);
  Long_Int(const char *src);
  long _int;
};


class String : public Value
{
 public:
  String ();
  static String *make(const char* string);
  static String *make(const String *string);

  ~String () { delete[] _data; }
  
  String *put (const char* src); // In some versions of LabBase
				 // the String with the put value
				 // is returned by put!
  char get (const int index) const;
  char* get () const;		// return entire string (null terminated)
  int high () const { return _count - 2; }
  void print(ostream&) const;
  String *save(P_SEGMENT *dbs);
  int compare(const Value&) const;
  String* substring(int start, int length) const;
  static String* zero_of();
  SLOT(unsigned short,count);	// Number of characters in string (including '\0');
  // Put this member first to save space!

protected:
  void get_error (const int index) const;
  char* _data;  // '\0'-terminated !!!
  String (const char* string);
  String (const String* string);
};


class Sequence : public String
{
public:
  static Sequence *make(const char*);
  static Sequence *make(const  String*);
  Sequence* reverse_complement() const;
  Sequence* reverse_complement(int, int) const;
  static char reverse_complement(char);
  int valid() const;
  static Sequence *zero_of();
protected:
  Sequence(const char*);
  Sequence(const String*);
};


class Float : public Value
{
public:
  static Float *make(const char* src);
  static Float *make(double);
  operator double () const { return _float; }
  void print(ostream&) const;
  Float *save(P_SEGMENT *dbs);
  int compare(const Value&) const;
  static Float *zero_of();
protected:
  double _float;
  Float (const char* src);
  Float (double);
};


class Date : public Value
{
public:
  static Date* make(time_t clock);
  static Date* make(const char* s);

  struct tm* time_as_struct () const { return localtime (&_clock); }
  struct tm* time_as_struct (struct tm* tm)
    { _clock = timelocal (tm);	
      return tm; }
  time_t time_in_seconds () const
    { return _clock; }
  time_t time_in_seconds (time_t clock)
    { _clock = clock; return _clock; }
  void print(ostream&) const;
  Date *save(P_SEGMENT *dbs);  // Do not use!
  int compare(const Value&) const;
  static Date *zero_of();
protected:
  time_t _clock;
  Date(const char* s);
  Date (time_t clock) : Value(DATE) { time_in_seconds (clock); }
};



#ifdef OBJECTSTORE
///////////////////////////////////////////////////////////////////////////////
// CLASS COLLECTION  
// !!!!!!!!  DO NOT USE THESE CLASSES (Collection<T>, Set<T>, List<T>)
// !!!!!!!!  Outside of value.{C,H}  !!!!!!!!!!!!!!!!!!!!!!!!
// !!!!!!!!  To do so will introduce dependencies on objectstore
// !!!!!!!!  that we can do without!
//
// Collections
//
// INHERENT LIMITATION: class T MUST BE a pointer type.  This class will
//          FAIL HORRIBLY unless this is the case
//
template<class T> class Collection : public Value
{
public:
  Collection () : Value(UNDEFINED) {
    os_collection::initialize();
    _collection = &os_Collection<T>::create(os_segment::of(this));
  }
  
  Collection (int behavior,
	      int expected_size = 0) : Value(UNDEFINED) {
		os_collection::initialize (); // This is a no-op if called a second time
		_collection = &os_Collection<T>::create(os_segment::of(this),
							behavior,
							expected_size);
	      }
  
  ~Collection () { delete _collection; }
  
  
  // os_cursor::safe, in association with the os_collection::maintain
  //_cursors option of the collection ensures that the cursor remains
  // valid (i.e._is not corrupted_) if an element is deleted at the
  // cursor position.  This eventuallity is highly probable given
  // the lbback query language, since one might iterate through
  // a material set or sets of material sets, and delete 
  // elemenents that satisfy particular conditions.
  
  // Return an os_cursor
  os_Cursor<T>* os_cursor (os_int32 safety=os_cursor::safe) const {
    return new os_Cursor<T>(*_collection,safety);
  }
  
  os_Cursor<T>* os_cursor (const os_index_path &path,
			   os_int32 safety=os_cursor::safe) const {
			     return new os_Cursor<T>(*_collection,path,safety);
			   }
  
  // The following are pass-through methods
  // Add an ObjectStore index (for use with ObjectStore queries)
  void add_index(const os_index_path& path,
		 os_int32 options = os_index_path::unordered,
		 os_segment *segment = 0	// 0 means to use the same segment
		 // as the Collection.
		 ) {
    _collection->add_index(path, options, segment);
  }
  
  // The inverse of add_index
  void drop_index(const os_index_path& path) {
    _collection->drop_index(path);
  }
  
  // Use queries to use the indexes
  Collection<T> *query(char *element_type_name, char *query_string,
		       os_database *schema_db=0) {
    Collection<T> *r = new Collection<T>; // Deliberately transient.
    delete r->_collection;
    r->_collection 
      = &_collection->query(element_type_name, query_string, schema_db);
    return r;
  }
  
  T query_pick(const os_bound_query& q) { return _collection->query_pick(q); }
  
  // This version of query_pick seems to be associated with
  // with a storage management bug in ObjectStore.
  // Use query_pick(os_bound_query&) instead.
  //
  // T query_pick(char *element_type_name, char *query_string,
  //		 os_database *schema_db=0) {
  //  return
  // _collection->query_pick(element_type_name, query_string, schema_db);
  // }
  
  int cardinality() const { 
    return _collection->cardinality(); 
  }
  
  void change_behavior(int behavior) { 
    _collection->change_behavior(behavior); 
  }
  
  void clear () { 
    _collection->clear(); 
  }
  
  int contains (T element) const { 
    return 0 != _collection->contains(element); 
  }
  
  int empty () const { 
    return 0 == cardinality(); 
  }
  
  int get_behavior () const { 
    return _collection->get_behavior(); 
  }
  
  void insert (T element) {
    _collection->insert(element);
  }
  
  T pick () const { 
    return _collection->pick(); 
  }
  
  void put (T element) { insert(element); }
  
  int remove (T element) {
    return _collection->remove(element);
  }
  
  operator os_Collection<T>& () { return *_collection; }
  
protected:
  os_Collection<T>* _collection;
};


template<class T> class Set : public Collection<T>
{
public:
  Set () 
    :  Collection<T>(os_collection::maintain_cursors) { _type = (char) SET; }
  Set (int expected_size)
    : Collection<T>(os_collection::maintain_cursors,expected_size)
      { _type = (char) SET; }
};
#endif

class QL_Args;


class V_Cursor  // Cursor over a V_Collection
{
public:
  virtual ~V_Cursor();
  virtual Value *first() = 0;
  virtual Value *next() = 0;
};


class V_Collection : public Value
{
public:
  static V_Collection *make(Value::tv_type_id, const V_Collection*c,
			    Value* zero_Value);
  // For SLIST only.

  static V_Collection *make(Value::tv_type_id, const QL_Args *args=0);
  // For SET, LIST, and TUPLE only.

  static V_Collection *make(Value::tv_type_id, const List_Value *);
  static V_Collection *make(Value::tv_type_id, unsigned nelems,Value*[]);

  // Declare cursor classes as friends to allow them access to _data member.
  friend class V_Cursor_STD;
  friend class V_Cursor_Sparse;

  ~V_Collection() { delete[] _data;}

  V_Collection* save(P_SEGMENT * db);
  void unsave();
  void print(ostream&) const;
  int compare(const Value&) const;
  Value *get(int) const; 
  V_Cursor *cursor_of() const;
  int cardinality() const;
  static V_Collection *set_zero_of();
  static V_Collection *list_zero_of();
  static V_Collection *slist_zero_of();

protected:
  SLOT(unsigned short,count);	// Number of elements in the collection.
  Value** _data;
  V_Collection(tv_type_id, int size=0);
  void put(int index, Value *v);
};

Listdeclare(Value);
#endif				// last line of file
